//
//  TDCDriver.h  (W.Badgett)
//
//  Class to configure and readout an optical chain 
//  of Sten Henson's TDC 
//

#ifndef _TDCDriver_h
#define _TDCDriver_h

#include <stdio.h>
#include <fstream>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "TDCFragment.h"
#include "LariatConfiguration.h"

class TDCDriver
{
  public:
  enum
  {
    /* MAX_TDCS = 16, */
    /* MAX_CHAMBERS = 4, */
    /* N_CHANNELS = 64, */
    /* TDCS_PER_CHAMBER = 4, */
    CSR_ADDR = 0x1,
    START_SPILL_ADDR = 0x2,
    END_SPILL_ADDR = 0x3,
    SPILL_TRIGGER_COUNTER_HIGH_ADDR = 0x4,
    SPILL_TRIGGER_COUNTER_LOW_ADDR = 0x5,
    TDC_SPILL_STATUS_ADDR = 0x6,
    SPILL_COUNTER_ADDR = 0x7,
    HEADER_BUFFER_POINTER_ADDR = 0x8,
    HEADER_BUFFER_DATA_PORT_ADDR = 0x9,
    HEADER_BUFFER_WRITE_POINTER_ADDR = 0xa,
    SPILL_STATE_ADDR = 0xb,
    LINK_ACTIVITY_ADDR = 0xc,
    SPILL_WORD_COUNT_HIGH_ADDR = 0xe,
    SPILL_WORD_COUNT_LOW_ADDR = 0xf,
    TDC_WORD_COUNT_BASE_ADDR = 0x10,
    TDC_PORT_LINK_STATUS_BASE_ADDR = 0x20,
    TDC_PORT_INTERRUPT_STATUS_BASE_ADDR = 0x30,
    PULSER_FREQUENCY_HIGH_ADDR = 0x7C,
    PULSER_FREQUENCY_LOW_ADDR = 0x7D,
    PULSER_SPILL_WIDTH_ADDR = 0x7E,
    PULSER_INTERSPILL_WIDTH_ADDR = 0x7F,
    TRIGGER_TIME_STAMP_ADDR = 0x89,
    SPILL_GATE_S = 3,
    EOT_COMPLETE_S = 2,
    TDC_READOUT_DONE_S = 1,
    TDC_READOUT_BUSY_S = 0,
    FIFO_BUFFER_RESET_S = 7,
    READOUT_SEQUENCER_RESET_S = 6,
    RESET_SPILL_COUNTER_S = 5,
    SINGLE_STEP_S = 4,
    TEST_PULSER_ENABLE_S = 3,
    SEND_2_FM_TX_ENABLE_S = 2,
    PLL_REFERENCE_SOURCE_S = 1,
    DMA_BUSY_S = 0,
    CONTROLLER_MEMORY_SIZE = 0x10000000,
    GOOD_TDC_STATUS = 0x0008,
    TDC_TDC_SPILL_STATUS_ADDR = 0x39,
    TDC_TDC_CSR00_NOMINAL = 0x7,
    TDC_TDC_IGNORE_SPILL_S = 4,
    TDC_TDC_OUTPUT_SELECT_S = 6,
    TDC_TDC_CSR_ADDR = 0,
    TDC_TDC_GATE_WIDTH_ADDR = 1,
    TDC_TDC_INITIAL_TIMESTAMP_ADDR = 2,
    TDC_TDC_PIPELINE_DELAY_ADDR = 3,
    TDC_TDC_CHANNEL_ENABLE_BASE_ADDR = 4,
    TDC_TDC_THRESHOLD_BASE_ADDR = 8,
    TDC_TDC_INITIAL_TIMESTAMP_V = 0xAC,
    TDC_TDC_ALL_CHANNELS = 0xFFFF
   };


  bool ok;
  LariatConfiguration * config;
  int localSocket;
  struct sockaddr_in remoteSocket;
  struct hostent *remoteHost;
  char addressString[256];
  uint16_t * buffer;

  uint16_t csr;
  uint16_t tclkStart;
  uint16_t tclkEnd;
  uint32_t triggerCounter;
  uint16_t tdcSpillStatus;
  uint16_t spillCounter;
  uint16_t spillState;
  uint16_t spillWidth;
  uint16_t tdcInterruptStatus[TDCFragment::MAX_TDCS];
  uint16_t tdcLinkStatus[TDCFragment::MAX_TDCS];
  uint16_t tdcSpill[TDCFragment::MAX_TDCS];
  uint16_t interSpillWidth;
  uint32_t spillWordCount;
  uint32_t pulserFrequency;

  bool spillGate;
  bool eotComplete;
  bool tdcReadoutDone;
  bool tdcReadoutBusy;
  bool dmaBusy;
  bool pllReferenceSource;
  bool testPulserEnabled;
  bool singleStep;

  uint16_t tdcEnabled;
  uint16_t tdcWordCount[TDCFragment::MAX_TDCS];
  uint16_t tdcLink[TDCFragment::MAX_TDCS];
  uint16_t tdcInterrupt[TDCFragment::MAX_TDCS];

  ~TDCDriver();
  TDCDriver(LariatConfiguration * config);
  void print(std::ostream & os = std::cout);
  void readEvent(std::ostream *dataFile);

  void write(uint32_t address, uint32_t data, int32_t tdc=-1);
  uint32_t read(uint32_t address, int32_t tdc=-1);

  void readConfiguration();
  bool spill();
  uint32_t getSpillWordCount();
  int countBits(uint16_t mask);
  int minBit(uint16_t mask);
  int maxBit(uint16_t mask);
  bool checkTdcNumber(uint8_t n, uint8_t expected);
  char * zeit();

 protected:
  int ptr, eptr;
  uint32_t dataSize, eventSize;
  void incr()
  { if ( ptr < static_cast<int>(dataSize) ) { ptr++;} }
  void incre()
  {  if ( ( ptr < static_cast<int>(dataSize)) && ( eptr < static_cast<int>(eventSize) )) { ptr++; eptr++;} }
};


#endif
