#ifndef V1495_DRIVER_H
#define V1495_DRIVER_H

#include "CAENBase.h"
#include <stdint.h>
#include "LariatConfiguration.h"
#include "V1495Configuration.h"
#include "V1495Fragment.h"
#include <iostream>

#ifdef USEVMELIB
extern "C"
{
#include "CAENVMElib.h"
}
#include "CAENVMEtypes.h"
#else
extern "C"
{
#include "CAENComm.h"
}
#endif

/** 
 * This class represents the V1495 logic module.
 * It relies on the CAENComm library.
 * Functions in this class use the CAENComm error checking 
 * and throw exceptions of type caen_exception.
 * OR
 * It uses the CAEMVMElib because of some stupid intransigence on
 * the part of CAEN. If you open a handle to the V2718 with CAENVMElib
 * then you cannot also use CAENComm. But you need CAENVMLlib to configure
 * the 2718 pulsers, etc. But CAENComm has higher level functionality, 
 * better error reporting, etc. Sigh. Why can't CAEN get their act together??
 * 
 * NOTE: this situation is fixed in new versions of the CAEN libraries.
 */
class V1495Driver {
 public:
  /**
   *  Construct the driver based on the configuration objects.
   *  The caen::vme_handle is only needed if we are using the VMElib
   *  but, in that case it has to be externally supplied.
   */
  V1495Driver(const LariatConfiguration* lariat, V1495Configuration* config, caen::vme_handle h=0);


#ifdef USEVMELIB

  V1495Driver(caen::vme_handle h, uint32_t base_address);

#else
  V1495Driver(uint32_t base_address, int linknum=0, int board=0,
	    CAENComm_ConnectionType linktype=CAENComm_OpticalLink);
#endif

  virtual ~V1495Driver();


  /**
   * print the configuration and most recent event data in
   * a human readable format.
   */   
  virtual void print(std::ostream& os = std::cout);
  //  virtual void readConfiguration(){}

  /**
   * Configure the board using V1495Configuration* config;
   */
  virtual void configureCommon();
  virtual void configurePedestal(){configureCommon();}
  virtual void configurePhysics(){configureCommon();}

  /**
   * Readout the V1495 storing the result in V1495Fragment frag.
   */
  virtual void readEvent();

  /**
   * Readout the V1495 storing the result in V1495Fragment frag
   * and writing the fragment to the output stream.
   */
  virtual void readEvent(std::ostream * os);


  /**
   * Pulls the version of the VME FPGA firmware.
   * This is distinct from the USER firmware that we program.
   */
  void fwVersion(uint32_t& X, uint32_t& Y);
  
  /**
   * Issues a board reset. This translates into the nLBRES signal.
   * That will zero many of the counters on the board.
   */
  void reset();

  /**
   * Reloads the user FPGA firmware from the flash memory.
   */ 
  void reloadFirmware();

  /**
   * This does a 32 bit read of the selected register.
   * The register address is relative to baddr.
   * (i.e., this function adds baddr to reg).
   */
  uint32_t readRegister(uint32_t reg);

  /**
   * This does a 32 bit write of the selected register.
   * The register address is relative to baddr.
   * (i.e., this function adds baddr to reg).
   */
  void writeRegister(uint32_t reg, uint32_t data);

  /**
   * Gives back the VME handle used by CAENComm if we are using CAENComm.
   * Or, directly pass back the handle if we are using CAENVMELib.
   * The handle can be used to call functions in CAENVMELib
   */
  caen::vme_handle getVMEHandle();

  V1495Fragment& getFragment(){return frag;}

 private:
  void configureControlRegister();
  void configureDelays();
  void configureCoincidence(const V1495TriggerPattern& );
  


#ifdef USEVMELIB
  caen::vme_handle h;
  
#else 
  caen::comm_handle h;
#endif

  uint32_t baddr;
  uint32_t fwX;
  uint32_t fwY;

  V1495Configuration* config;  
  V1495Fragment frag; //!< the most recent data read-out from the V1495

};

#endif 
