#ifndef V2718_DRIVER_H
#define V2718_DRIVER_H

#include "CAENBase.h"
#include <stdint.h>
#include <iostream>
#include <unistd.h>
#include "LariatConfiguration.h"
#include "V2718Configuration.h"
#include "CAENVMElib.h"
#include "CAENVMEtypes.h"



/** 
 * This class represents the V2718 crate controller.
 * It uses CAENVMELib under the covers.
 * Most communication with modules (i.e., V1495,...) is
 * done using the CAENComm or CAENDigitizer libraries, and in fact,
 * in order play nicely with them, we need to get the vme handle
 * from their board info structure and pass it into the constructor below.
 */
class V2718Driver {
 public:
  /**
   *  Construct the driver based on the configuration objects.
   *  The caen::vme_handle is only needed in order to use the 
   *  CAENVMELib, necessary to talk to the V2718 controller itself.
   */
  V2718Driver(const LariatConfiguration* lariat, 
	      V2718Configuration* config, caen::vme_handle h);


  virtual ~V2718Driver();

  /**
   * printout some information about the V2718
   */ 
  virtual void print(std::ostream& os = std::cout);

  /**
   * Configure the board using V2718Configuration* config;
   */
  virtual void configureCommon() {configureReadoutRegister();}
  virtual void configurePedestal(){configureCommon();}
  virtual void configurePhysics(){configureCommon();}

  /**
   * Configure the on-board pulser to output a long pulse
   * when input source 0 sees a pulse.
   * When the pulser is active a bit is set in a register on 
   * the V2718. The plan is to start the pulser with the
   * end of spill signal, and poll the register. This tells us
   * when to read out.
   */
  virtual void configureReadoutRegister();
  
  /**
   * Poll the readout register ntimes, each separated 
   * by time_sep in microseconds. If ntimes==1 time_sep does nothing.
   * Returns true if the register has been set high due to a pulse
   * on input source 0.
   */
  virtual bool pollReadoutRegister(int ntimes=1, useconds_t time_sep=10000);

  /*
   *  Read and return the spill flag once
   */
  virtual bool spill();

  /**
   * Reset the readout register. Done after 
   */
  virtual void resetReadoutRegister();

 private:
  void check(CVErrorCodes error_code, const char* place="");
  void check(const char* place="");  

  caen::vme_handle h;  
  V2718Configuration* config;  
  CVErrorCodes err;

};

#endif 
