#!/bin/env bash


if [ "$#" != 1 ]; then
    echo "Script expects one argument: the # of desired output Root files (1 file per spill assumed)"
    echo "Passing 0 as an argument means you want to cleanup a preexisting DAQ"
    return 1
fi

# Keep track of when this script began; we'll use this information
# when deciding which logfiles to list to the user at the end of
# running

let "starttime=`date +%s`"

# Please note that the number of requested root files will be assumed
# to be synonymous with the number of requested events; see the
# "manage1x1x2System.sh -n <nevents> stop" command below

nrootfiles=$1

# Unless isDebug is set to "0", debug (as opposed to profile)
# compilation of lariatonline will be used

isDebug="0"

if [ -z "$LARIATONLINE_DIR" ]; then
    echo "lariat-online does not appear to have been set up"
    return 101
fi

lariatartdaqbase=$LARIATONLINE_DIR/source/daq

# "currdir" is used so that the script, when sourced, can find its way
# back to the directory in which it was sourced

currdir=$PWD


# "disableTimeReport", when not set to "0", will allow users to strip
# out the timestamp provided by pmt.rb, reducing screen clutter; I'll
# need to think about this as it may result in a real loss of
# information, however

disableTimeReport="0"

# Directory in which log files from the transitions will be searched
# for, for purposes of error checking

transitionlogdir=/daqdata/log/masterControl

# Directory in which the log(s) of the pmt.rb will appear

pmtlogdir=/daqdata/log/pmt

# 10/8/14

# I'm trying to capture pmt.rb's stderr in a file which can be
# selectively examined for error messages (as is done in the case of
# the transition commands thanks to the "logerrorcheck" function) but
# for now, at least, it keeps going to the screen (i.e., stdout)

pmterrlog="pmterrlog."`uuidgen`


# The "find_procs" function will return the process ID of a running
# process which matches the wildcard pattern passed to it in the "ps
# aux" command

# So, for example, if you wish to get the process ID for a running
# instance of DemoControl.rb (the script manage1x1x2System.sh wraps
# around):

# find_procs "ruby.*DemoControl.rb"

# Or for the process management tool:

# find_procs "pmt.rb"

# In the event that more than one such process is running,
# "find_procs" will return the PID of the first process it finds. If
# no such process is running, it will return 0.

function find_procs {

    cmd_seek_process='ps aux | grep "$1" | grep "'$USER'" | grep -v grep'
    nprocs=`eval $cmd_seek_process | wc -l`

    if [ "$nprocs" -eq "0" ]; then
	runningid="0"
    else
	runningid=`eval $cmd_seek_process | head -1 | awk '{print $2}' `
    fi

    echo $runningid
}

# full_cleanup is designed to kill off all DAQ-related processes; it
# should be called at the end of this script as well as used as a
# handler for ctrl-C.

# Should full_cleanup discover that its efforts to kill a process with
# a standard "kill" command isn't working, it will resort to "kill
# -9", and then return "0" (for "not clean exit"). However, pmt.rb
# appears to always need a "-9", so in this case the exit is still
# considered clean


function full_cleanup {

    echo "EXECUTING FULL PROCESS CLEANUP"
    let "cleanexit=1"
    
    for wildcard in  "ruby.*DemoControl.rb" "pmt.rb"; do

	local procid="1000000"
	let "tries=0"
	
	while [ "1" ] ; do

	    let "lastprocid=$procid"
	    procid=$(find_procs $wildcard)
	    
	    if [ "$procid" -eq "$lastprocid" ] ; then
		let "tries=tries+1"
	    else
		let "tries=0"
	    fi


	    if [ "$procid" != "0" ] && ((tries >= 5)); then
		echo "kill -9 $procid"
		kill -9 $procid

		let "cleanexit=0"

	    elif [ "$procid" != "0" ]; then
		echo "kill $procid"
		kill $procid

		# Process management tool takes a few seconds to shut down

		if [ "$wildcard" == "pmt.rb" ] ; then
		    sleep 3
		fi
	    else
		break
	    fi

	done
    done

    # In case alterations to the terminal were made (e.g., so the
    # "read" command would timeout while waiting for a keyboard hit),
    # return it to normal

    if [ -t 0 ]; then
	stty sane
    else
	echo "Currently unable to determine the stdin file descriptor; please contact John Freeman"
    fi

    # Only remove the file containing pmt.rb's stderr output if it's empty

    if ! [ -s $pmterrlog ] ; then
	rm -f $pmterrlog
    fi

    if (( cleanexit == 1 )) ; then
	echo "CLEAN EXIT: ALL PROCESSES NORMALLY KILLED" 
    else
	echo "UNCLEAN EXIT: AT LEAST ONE PROCESS FORCIBLY KILLED"
    fi

    echo "For more information, examine the following logfiles: "

    let "endtime=`date +%s`"
    let "minutesrunning = (endtime - starttime)/60 + 1"

    find $pmtlogdir -type f -name "*.log" -mmin -${minutesrunning} | sort
    find $transitionlogdir -type f -name "*.log" -mmin -${minutesrunning} | sort

    # And finally, return to the directory whence this script was called
    cd $currdir
}

## trap ctrl-c and call full_cleanup()
trap 'echo "\"CTRL-C\" APPEARS TO HAVE BEEN HIT; WILL PERFORM HARD EXIT" >&2 ; full_cleanup; return' INT


# "persistent_stop" will basically keep issuing the "stop" transition
# until it sees that no processes return "Busy" (either because
# they've returned to the ready state, or because they no longer
# exist). If any processes do return "Busy", this function will
# reissue "stop"

function persistent_stop {

    waittime=5

    while [ "1" ] ; do

	echo "Running stop..."
	manage1x1x2System.sh stop > /dev/null
	nstopped=`manage1x1x2System.sh status | grep "result: Busy" | wc -l`

	if [ "$nstopped" -eq "0" ] ; then
	    echo "Done"
	    break
	else
	    echo "At least one process is busy; will retry..." >&2
	fi
    
	sleep $waittime
    done
}



if [ "$nrootfiles" == "0" ] ; then

    echo "# of requested output files is 0; interpreting this as cleanup request..."
    
    full_cleanup
    return 0 
fi

if ! [ -d $lariatartdaqbase ] ; then
    echo "Unable to find expected Lariat code base directory $lariatartdaqbase" >&2
    return 10
fi


export LARIATARTDAQ_DIR=$lariatartdaqbase/lariat-artdaq

if [ ! -d $LARIATARTDAQ_DIR ] ; then
    echo "Unable to locate directory $LARIATARTDAQ_DIR"
    return 100
fi

cd $lariatartdaqbase/lariat-artdaq/tools

source $lariatartdaqbase/lariat-artdaq/tools/setupDemoEnvironment.sh ""   

# JCF, 12/12/14

# The "manage1x1x2System.sh" script already has a definition for where
# the keypress file is located; the following lines are a bit awkward
# but this ensures there's no mismatch between the keypress file used
# in the manage script and the keypress file used in this script

m112ss=`which manage1x1x2System.sh` 
keypress_file=`cat $m112ss | sed -r -n 's/keypress_file="(.*)"/\1/p'`

if [ ! -f $keypress_file ] ; then
    echo "Problem extracting keypress filename from $m112ss" >&2
    return 200
fi


# create the configuration file for PMT                                                                   
pmtFile="/daqdata/log/pmt/pmtConfig.$$"
set +C ; > $pmtFile ; set -C

if [ ! -f $pmtFile ] ; then
    echo "Problem opening up $pmtFile" >&2
    return 3
fi 

echo "BoardReaderMain `hostname` ${LARIATARTDAQ_BR_PORT[0]}" >> $pmtFile
echo "EventBuilderMain `hostname` ${LARIATARTDAQ_EB_PORT[0]}" >> $pmtFile
echo "AggregatorMain `hostname` ${LARIATARTDAQ_AG_PORT[0]}" >> $pmtFile
echo "AggregatorMain `hostname` ${LARIATARTDAQ_AG_PORT[1]}" >> $pmtFile

echo "Starting up the Process Management Tool, pmt.rb"


if [ "$disableTimeReport" != "0" ] ; then
    pmt.rb -p $LARIATARTDAQ_PMT_PORT -d $pmtFile --logpath /daqdata/log --display $DISPLAY | sed -r -e 's/^.*20[[:digit:]][[:digit:]]\:\s*//' 2>$pmterrlog  &
else
    pmt.rb -p $LARIATARTDAQ_PMT_PORT -d $pmtFile --logpath /daqdata/log --display $DISPLAY 2>$pmterrlog  & 
fi

# Wait a few seconds between when the artdaq processes are fired up
# and when you send them the init command

sleep 5

# "logerrorcheck" is a useful but fairly complex function designed to
# display only a desired subset of the error messages sent
# stdout/stderr during a transition, as well as to set a variable of
# the user's choosing designed to indicate whether or not any such
# error occurred. Example uses can be found below; to take one
# example:

# logerrorcheck okstatus "init" "1" "Connection refused" "Exception"

# So what happens here is the following

# In the first argument, the user supplies the (pre-existing) okstatus
# variable, which logerrorcheck will set to "0" if an error message
# (or messages) is found, and to "1" if everything's fine. 

# In the second, the user supplies the name of the command (this
# can be "init", "start", "stop", "shutdown" or "status")

# The third, if set to "0", will cause logerrorcheck to print out
# which logfile it's checking for errors (useful if called once, messy
# if called within a loop)

# The remaining arguments are tokens grep'd for in the logfile. So in
# the example above, any lines with either "Connection refused" or
# "Exception" will be considered to be reporting errors, which will be
# relayed to stdout and which will cause okstatus, above, to be set to
# "0"

function logerrorcheck {

    local __allOKVarname=$1

    # Get the most recent instance of the log for this transition
    local log=$( ls -tr1 $transitionlogdir/*${2}.log | tail -1 )

    if [[ "$log" != "" ]] && [[ "$3" != "0" ]]; then
	echo "Checking $log for errors"
    elif [[ "$log" == "" ]] ; then
	echo "Error: Couldn't find a log of the form *$1.log in $transitionlogdir!!" >&2
	return
    fi


    shift # Pop off the name of the global variable to set
    shift  # And the transition token 
    shift # And the "print-the-logfile-name" boolean

    # And now look at the "error message" tokens

    local allOK="1"

    for tok in "$@" ; do
	res=$( grep -Hn "$tok" $log )
	
	if [ "$res" != "" ]; then
	    echo $res >&2
	    allOK="0"
	fi
    done

    eval $__allOKVarname="'$allOK'"
}


echo "Running init..."
manage1x1x2System.sh init > /dev/null
echo "Done"

okstatus="999"
logerrorcheck okstatus "init" "1" "Connection refused" "Exception" "Error initializing" "not found" "execution expired" "exception"

if [ "$okstatus" == "0" ]; then
    full_cleanup
    echo "Problem found during initialization" >&2
    return
fi




echo "Running start..."
manage1x1x2System.sh start > /dev/null
echo "Done"
logerrorcheck okstatus "start" "1" "Connection refused" "Exception" "exception"

if [ "$okstatus" == "0" ]; then
    full_cleanup
    echo "Problem found during start" >&2
    return
fi


# Run "stop" in the background so that if the user wishes to stop
# running before $nrootfiles output files have been created, this
# script can register the "q"

echo "Running stop w/ an expectation of $nrootfiles spills..."
manage1x1x2System.sh -n $nrootfiles stop > /dev/null &


let "looptimesecs=1"
let "looptimetenths=((10*looptimesecs))"
keypress=''

if [ -t 0 ]; then 
    stty -echo -icanon time $looptimetenths min 0; 
else
    echo "Currently unable to determine the stdin file descriptor-- potential problems could arise with registering keystrokes; please contact John Freeman" >&2
fi


# Loop either until someone hits the "q" key, or the DAQ process
# (manage1x1x2System.sh -n <nspills> stop) to be done

# In the body of the loop, check the output of the two processes
# running in the background (process management tool and the DAQ
# process) for errors


nprocs="1"
while ! [ "x$keypress" = "xq" ] && [ "$nprocs" != "0" ] ; do
  read keypress

  # JCF, 12/12/14

  # Also register "p" (print) and "h" (help), used in LariatReadout::readSpill() 

  if [[ x$keypress == xp ]] || [[ x$keypress == xh ]]; then
      set +C; echo -n "$keypress" > $keypress_file; set -C;
  fi


  logerrorcheck okstatus "stop" "0" "Connection refused" "Exception" "There was a problem communicating" "exception"

  if [ "$okstatus" == "0" ]; then
      full_cleanup
      echo "Problem found during running" >&2
      return
  fi

  if [ -s "$pmterrlog" ] ; then
      echo "Problem with pmt.rb ; please take a look at its stderr, stored in $pmterrlog"
      full_cleanup
      return 4
  fi


  nprocs=$( find_procs 'ruby.*DemoControl.rb' )
done

if [ "$keypress" == "q" ] ; then
    echo '"q" APPEARS TO HAVE BEEN HIT; WILL PERFORM SOFT EXIT' >&2
    persistent_stop
fi

echo "Shutting down..."
manage1x1x2System.sh shutdown > /dev/null
echo "Done"

cd $currdir

full_cleanup
return


