#!/bin/bash

# JCF, 6/16/14
# My edits to LariatReadout::readSpill ensure that the keypress file
# is checked for (among other things) the "q" character, which will be
# added on the "stop" transition

# JCF, 7/19/14
# Added in logic s.t. the keypress file is only affected in the case
# of reading from hardware

keypress_file="/opt/lariat-online/keypress.txt"

source `which setupDemoEnvironment.sh` ""

AGGREGATOR_NODE=`hostname`
THIS_NODE=`hostname -s`

# this function expects a number of arguments:
#  1) the DAQ command to be sent
#  2) whether to run online monitoring
#  3) the data directory
#  4) the logfile name
#  5) whether to write data to disk [0,1]
#  6) the desired number of events in the run (stop command)
#  7) the desired duration of the run (minutes, stop command)
#  8) the desired size of each data file
#  9) the desired number of events in each file
# 10) the desired time duration of each file (minutes)
# 11) whether to print out CFG information (verbose)
function launch() {

  enableSerial=""
  if [[ "${12}" == "1" ]]; then
      enableSerial="-e"
  fi

 # JCF, 7/15/14

 # Run number will be hardwired as 99,999,999 if we're not in hardware
 # mode, otherwise get it from /home/nfs/lariat/config/runNumber.dat

  # Set "hardwareMode" either to "0" or to "1"

  #hardwareMode=0
  hardwareMode=1

  runNumber=""
  boardreaderFile=""

  if [ "$hardwareMode" == "0" ] ; then

      runNumber="99999999"
      boardreaderFile="$LARIATONLINE_DIR/fcl/BoardReader_for_Lariat_nohardware_1x1x2.fcl"

  elif [ "$hardwareMode" == "1" ] ; then
      
      if [ ! -e /home/nfs/lariat/config/runNumber.dat ] ; then
	  echo "Error in $0: unable to find /home/nfs/lariat/config/runNumber.dat"
      fi

      runNumber=`cat /home/nfs/lariat/config/runNumber.dat`
      boardreaderFile="$LARIATONLINE_DIR/fcl/BoardReader_for_Lariat_hardware_1x1x2.fcl"

  else

      echo "Error in $0: hardwareMode value not allowed to be "'"'$hardwareMode'"'
      exit 1

  fi

  DemoControl.rb ${enableSerial} -s -c $1 \
    --toy1 `hostname`,${LARIATARTDAQ_BR_PORT[0]},0 \
    --eb `hostname`,${LARIATARTDAQ_EB_PORT[0]} \
    --ag `hostname`,${LARIATARTDAQ_AG_PORT[0]},1 \
    --ag `hostname`,${LARIATARTDAQ_AG_PORT[1]},1 \
    --brfcl $boardreaderFile \
    --ebfcl $LARIATONLINE_DIR/fcl/EventBuilder_for_Lariat_1x1x2.fcl \
    --agfcl $LARIATONLINE_DIR/fcl/Aggregator1_for_Lariat_1x1x2.fcl \
    --agfcl $LARIATONLINE_DIR/fcl/Aggregator2_for_Lariat_1x1x2.fcl \
    --data-dir ${3} --online-monitoring $2 \
    --write-data ${5} --run-event-count ${6} \
    --run-duration ${7} --file-size ${8} \
    --file-event-count ${9} --file-duration ${10} \
    --run-number $runNumber 2>&1 | tee -a ${4}
}

scriptName=`basename $0`
usage () {
    echo "
Usage: ${scriptName} [options] <command>
Where command is one of:
  help, config, init, start, pause, resume, stop, status, get-legal-commands,
  shutdown, start-system, restart, reinit, exit,
  fast-shutdown, fast-restart, fast-reinit, or fast-exit
General options:
  -h, --help: prints this usage message
Configuration options (init commands):
  -m <on|off>: specifies whether to run online monitoring [default=off]
  -D : disables the writing of data to disk
  -s <file size>: specifies the size threshold for closing data files (in MB)
      [default is 8000 MB (~7.8 GB); zero means that there is no file size limit]
  --file-events <count>: specifies the desired number of events in each file
      [default=0, which means no event count limit for files]
  --file-duration <duration>: specifies the desired duration of each file (minutes)
      [default=0, which means no duration limit for files]
  -o <data dir>: specifies the directory for data files [default=/daqdata]
End-run options (stop command):
  -n, --run-events <count>: specifies the desired number of events in the run
      [default=0, which ends the run immediately, independent of the number of events]
  -d, --run-duration <duration>: specifies the desired length of the run (minutes)
      [default=0, which ends the run immediately, independent of the duration of the run]
Notes:
  The primary commands are the following:
   * help - show the help from within the Lariat code
   * config - show the configuration of the Lariat hardware
   * init - initializes (configures) the DAQ processes
   * start - starts a run
   * pause - pauses the run
   * resume - resumes the run
   * stop - stops the run
   * status - checks the status of each DAQ process
   * get-legal-commands - fetches the legal commands from each DAQ process
  Additional commands include:
   * shutdown - resets the DAQ processes to their ground state (if
       needed), and stops the MPI program (DAQ processes)
   * start-system - starts the MPI program (the DAQ processes)
   * restart - this is the same as a shutdown followed by a start-system
   * reinit - this is the same as a shutdown followed by a start-system and an init
   * exit - this resets the DAQ processes to their ground state, stops the MPI
       program, and exits PMT.
  Expert-level commands:
   * fast-shutdown - stops the MPI program (all DAQ processes) no matter what
       state they are in. This could have bad consequences if a run is going!
   * fast-restart - this is the same as a fast-shutdown followed by a start-system
   * fast-reinit - this is the same as a fast-shutdown followed by a start-system
       and an init
   * fast-exit - this stops the MPI program, and exits PMT.
Examples: ${scriptName} -p 32768 init
          ${scriptName} -N 101 start
" >&2
}

# parse the command-line options
originalCommand="$0 $*"
onmonEnable=off
diskWriting=1
dataDir="/daqdata"
runEventCount=0
runDuration=0
fileSize=8000
fsChoiceSpecified=0
fileEventCount=0
fileDuration=0
verbose=0
OPTIND=1
while getopts "hc:o:t:m:Dn:d:s:w:v-:" opt; do

    if [ "$opt" = "-" ]; then
        opt=$OPTARG
    fi
    case $opt in
        h | help)
            usage
            exit 1
            ;;
        m)
            onmonEnable=${OPTARG}
            ;;
        o)
            dataDir=${OPTARG}
            ;;
        D)
            diskWriting=0
            ;;
        n)
            runEventCount=${OPTARG}
            ;;
        d)
            runDuration=${OPTARG}
            ;;
        run-events)
            runEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        run-duration)
            runDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-events)
            fileEventCount=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        file-duration)
            fileDuration=${!OPTIND}
            let OPTIND=$OPTIND+1
            ;;
        s)
            fileSize=${OPTARG}
            fsChoiceSpecified=1
            ;;
        v)
            verbose=1
            ;;
        *)
            usage
            exit 1
            ;;
    esac
done
shift $(($OPTIND - 1))

# fetch the command to run
if [ $# -lt 1 ]; then
    usage
    exit
fi
command=$1
shift

# verify that the command is one that we expect
if [[ "$command" != "start-system" ]] && \
   [[ "$command" != "help" ]] && \
   [[ "$command" != "config" ]] && \
   [[ "$command" != "init" ]] && \
   [[ "$command" != "start" ]] && \
   [[ "$command" != "pause" ]] && \
   [[ "$command" != "resume" ]] && \
   [[ "$command" != "stop" ]] && \
   [[ "$command" != "status" ]] && \
   [[ "$command" != "get-legal-commands" ]] && \
   [[ "$command" != "shutdown" ]] && \
   [[ "$command" != "fast-shutdown" ]] && \
   [[ "$command" != "restart" ]] && \
   [[ "$command" != "fast-restart" ]] && \
   [[ "$command" != "reinit" ]] && \
   [[ "$command" != "fast-reinit" ]] && \
   [[ "$command" != "exit" ]] && \
   [[ "$command" != "fast-exit" ]]; then
    echo "Invalid command."
    usage
    exit
fi

# translate the onmon enable flag
if [[ "$onmonEnable" == "on" ]]; then
    onmonEnable=1
else
    onmonEnable=0
fi

# JCF, 6/16/14
# See comment at top of script concerning the keypress file

# verify that we don't have both a number of events and a time limit
# for stopping a run
if [[ "$command" == "stop" ]]; then

# JCF, 10/4/14

# Only send "q" to the keypress file if the user wishes to exit
# instantaneously, i.e., if no event count or run duration argument is
# supplied to the "stop" transition

# #    if [[ "$hardwareMode" == "1" ]]; then

    if [[ "$runEventCount" == "0" ]] && [[ "$runDuration" == "0" ]]; then
	set +C; echo -n "q" > $keypress_file; set -C;
    fi
# #    fi

    if [[ "$runEventCount" != "0" ]] && [[ "$runDuration" != "0" ]]; then
        echo ""
        echo "*** Only one of event count or run duration may be specified."
        usage
        exit
    fi
fi

#if [[ "$command" == "pause" ]] && [[ "$hardwareMode" == "1" ]]; then

#    set +C; echo -n "q" > $keypress_file ; set -C

#fi

# In some instances (for example, something went wrong during the
# previous DAQ run), the keypress file may be non-empty at the start
# of a run; this block of code ensures that the file is cleared before
# datataking commences

if [[ "$command" == "start" ]]; then

    set +C; echo -n "" > $keypress_file ; set -C;

fi



if [[ "$command" == "help" ]] && [[ "$hardwareMode" == "1" ]]; then

    set +C; echo -n "h" > $keypress_file ; set -C;
    exit 0
fi


if [[ "$command" == "config" ]] && [[ "$hardwareMode" == "1" ]]; then

    set +C; echo -n "c" > $keypress_file ; set -C;
    exit 0
fi


# build the logfile name
TIMESTAMP=`date '+%Y%m%d%H%M%S'`
logFile="/daqdata/log/masterControl/lariatMC-${TIMESTAMP}-${command}.log"
echo "${originalCommand}" > $logFile
echo ">>> ${originalCommand} (Disk writing is ${diskWriting})"

# calculate the shmkey that should be checked
let shmKey=1078394880+${LARIATARTDAQ_PMT_PORT}
shmKeyString=`printf "0x%x" ${shmKey}`

# invoke the requested command
if [[ "$command" == "shutdown" ]]; then
    # # first send a stop command to end the run (in case it is needed)
    # launch "stop" $onmonEnable $dataDir \
    #     $logFile $diskWriting $runEventCount $runDuration $fileSize \
    #     $fileEventCount $fileDuration $verbose

    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "start-system" ]]; then
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.startSystem
elif [[ "$command" == "restart" ]]; then
    # # first send a stop command to end the run (in case it is needed)
    # launch "stop" $onmonEnable $dataDir \
    #     $logFile $diskWriting $runEventCount $runDuration $fileSize \
    #     $fileEventCount $fileDuration $verbose

    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.startSystem
elif [[ "$command" == "reinit" ]]; then
    # # first send a stop command to end the run (in case it is needed)
    # launch "stop" $onmonEnable $dataDir \
    #     $logFile $diskWriting $runEventCount $runDuration $fileSize \
    #     $fileEventCount $fileDuration $verbose

    # next send a shutdown command to move the processes to their ground state
    launch "shutdown" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
    # stop the MPI program
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    # clean up any stale shared memory segment
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    # start the MPI program
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.startSystem
    # send the init command to re-initialize the system
    sleep 5
    launch "init" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
elif [[ "$command" == "exit" ]]; then
    launch "shutdown" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.exit
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

elif [[ "$command" == "fast-shutdown" ]]; then
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
elif [[ "$command" == "fast-restart" ]]; then
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.startSystem
elif [[ "$command" == "fast-reinit" ]]; then
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.startSystem
    sleep 5
    launch "init" $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
elif [[ "$command" == "fast-exit" ]]; then
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.stopSystem
    xmlrpc ${THIS_NODE}:${LARIATARTDAQ_PMT_PORT}/RPC2 pmt.exit
    ssh ${AGGREGATOR_NODE} "ipcs | grep ${shmKeyString} | awk '{print \$2}' | xargs ipcrm -m 2>/dev/null"

else
    launch $command $onmonEnable $dataDir \
        $logFile $diskWriting $runEventCount $runDuration $fileSize \
        $fileEventCount $fileDuration $verbose
fi
