// ======================================================================
//
// RootOutputLariat
//
// ======================================================================

#include "RootOutputLariat.h"

#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/IO/PostCloseFileRenamer.h"
#include "RootOutputFileLariat.h"
#include "art/Framework/Principal/EventPrincipal.h"
#include "art/Framework/Principal/RunPrincipal.h"
#include "art/Framework/Principal/SubRunPrincipal.h"
#include "art/Persistency/Provenance/FileFormatVersion.h"
#include "art/Utilities/Exception.h"
#include "art/Utilities/unique_filename.h"
#include "art/Utilities/parent_path.h"
#include "cetlib/container_algorithms.h"
#include "cpp0x/utility"
#include "fhiclcpp/ParameterSet.h"
#include "boost/system/error_code.hpp"
#include "boost/algorithm/string/predicate.hpp"
#include "boost/algorithm/string/replace.hpp"
#include "boost/filesystem.hpp"


#include "TBranchElement.h"
#include "TObjArray.h"
#include "TTree.h"

#include <iomanip>
#include <sstream>
#include <fstream>

using art::RootOutputLariat;
using fhicl::ParameterSet;
using std::string;

namespace art {

  RootOutputLariat::RootOutputLariat(ParameterSet const& ps)
  : OutputModule               ( ps )
  , selectedOutputItemList_    ( )
  , catalog_                   ( ps.get<string>("catalog", string()) )
  , maxFileSize_               ( ps.get<int>("maxSize", 0x7f000000) )
  , compressionLevel_          ( ps.get<int>("compressionLevel", 7) )
  , basketSize_                ( ps.get<int>("basketSize", 16384) )
  , splitLevel_                ( ps.get<int>("splitLevel", 99) )
  , treeMaxVirtualSize_        ( ps.get<int64_t>("treeMaxVirtualSize", -1) )
  , saveMemoryObjectThreshold_ ( ps.get<int64_t>("saveMemoryObjectThreshold", -1l) )
  , fastCloning_               ( ps.get<bool>("fastCloning", true)
                                 && wantAllEvents() )
  , dropMetaData_              ( DropNone )  // tentative: see the c'tor body
  , dropMetaDataForDroppedData_( ps.get<bool>( "dropMetaDataForDroppedData"
                                             , false) )
  , moduleLabel_               ( ps.get<string>("module_label")  )
  , inputFileCount_            ( 0 )
  , rootOutputFile_            ( )
  , fstats_                    ( moduleLabel_, processName() )
  , filePattern_               ( ps.get<string>("fileName") )
  , tmpDir_                    ( ps.get<string>("tmpDir",
                                                parent_path(filePattern_)) )
  , lastClosedFileName_        ( )
  {
    string dropMetaData(ps.get<string>("dropMetaData", string()));
    if (dropMetaData.empty())                 dropMetaData_ = DropNone;
    else if (dropMetaData == string("NONE"))  dropMetaData_ = DropNone;
    else if (dropMetaData == string("PRIOR")) dropMetaData_ = DropPrior;
    else if (dropMetaData == string("ALL"))   dropMetaData_ = DropAll;
    else {
      throw art::Exception( errors::Configuration
                          , "Illegal dropMetaData parameter value: ")
          << dropMetaData << ".\n"
          << "Legal values are 'NONE', 'PRIOR', and 'ALL'.\n";
    }
  }  // c'tor

  RootOutputLariat::OutputItem::Sorter::Sorter(TTree * tree) {
    // Fill a map mapping branch names to an index specifying the order in the tree.
    if (tree != 0) {
      TObjArray * branches = tree->GetListOfBranches();
      for (int i = 0, sz = branches->GetEntries(); i != sz; ++i) {
        TBranchElement * br = (TBranchElement *)branches->At(i);
        treeMap_.insert(std::make_pair(string(br->GetName()), i));
      }
    }
  }

  bool
  RootOutputLariat::OutputItem::Sorter::operator()(OutputItem const& lh, OutputItem const& rh) const {
    // Provides a comparison for sorting branches according to the index values in treeMap_.
    // Branches not found are always put at the end (i.e. not found > found).
    if (treeMap_.empty()) return lh < rh;
    string const& lstring = lh.branchDescription_->branchName();
    string const& rstring = rh.branchDescription_->branchName();
    std::map<string, int>::const_iterator lit = treeMap_.find(lstring);
    std::map<string, int>::const_iterator rit = treeMap_.find(rstring);
    bool lfound = (lit != treeMap_.end());
    bool rfound = (rit != treeMap_.end());
    if (lfound && rfound) {
      return lit->second < rit->second;
    } else if (lfound) {
      return true;
    } else if (rfound) {
      return false;
    }
    return lh < rh;
  }

  void RootOutputLariat::fillSelectedItemList(BranchType branchType, TTree * theTree) {
    Selections const& keptVector =    keptProducts()[branchType];
    OutputItemList&   outputItemList = selectedOutputItemList_[branchType];
    for (auto const& selection : keptVector) outputItemList.push_back(OutputItem(selection));
    // Sort outputItemList to allow fast copying.  The branches in
    // outputItemList must be in the same order as in the input tree,
    // with all new branches at the end.
    cet::sort_all(outputItemList, OutputItem::Sorter(theTree));
  }

  void RootOutputLariat::openFile(FileBlock const& fb) {
    if (!isFileOpen()) {
      if (fb.tree() == 0) {
      fastCloning_ = false;
      }
      doOpenFile();
      respondToOpenInputFile(fb);
    }
  }

  void RootOutputLariat::respondToOpenInputFile(FileBlock const& fb) {
    for (int i = InEvent; i < NumBranchTypes; ++i) {
      BranchType branchType = static_cast<BranchType>(i);
      if (inputFileCount_ == 0) {
        TTree * theTree = (branchType == InEvent ? fb.tree() :
                          (branchType == InSubRun ? fb.subRunTree() :
                          fb.runTree()));
        fillSelectedItemList(branchType, theTree);
      }
    }
    ++inputFileCount_;
    if (isFileOpen()) {
      bool fastCloneThisOne = fb.tree() != 0 &&
                            (remainingEvents() < 0 || remainingEvents() >= fb.tree()->GetEntries());
      rootOutputFile_->beginInputFile(fb, fastCloneThisOne && fastCloning_);
      fstats_.recordInputFile(fb.fileName());
    }
  }

  void RootOutputLariat::respondToCloseInputFile(FileBlock const& fb) {
    if (rootOutputFile_) rootOutputFile_->respondToCloseInputFile(fb);
  }

  RootOutputLariat::~RootOutputLariat() {
  }

  void RootOutputLariat::write(EventPrincipal const& e) {
      if (hasNewlyDroppedBranch()[InEvent]) e.addToProcessHistory();
      rootOutputFile_->writeOne(e);
      fstats_.recordEvent(e.id());
  }

  void RootOutputLariat::writeSubRun(SubRunPrincipal const& sr) {
      if (hasNewlyDroppedBranch()[InSubRun]) sr.addToProcessHistory();
      rootOutputFile_->writeSubRun(sr);
      fstats_.recordSubRun(sr.id());
  }

  void RootOutputLariat::writeRun(RunPrincipal const& r) {
      if (hasNewlyDroppedBranch()[InRun]) r.addToProcessHistory();
      rootOutputFile_->writeRun(r);
      fstats_.recordRun(r.id());
  }

  // At some later date, we may move functionality from finishEndFile() to here.
  void RootOutputLariat::startEndFile() { }


  void RootOutputLariat::writeFileFormatVersion() { rootOutputFile_->writeFileFormatVersion(); }
  void RootOutputLariat::writeFileIndex() { rootOutputFile_->writeFileIndex(); }
  void RootOutputLariat::writeEventHistory() { rootOutputFile_->writeEventHistory(); }
  void RootOutputLariat::writeProcessConfigurationRegistry() { rootOutputFile_->writeProcessConfigurationRegistry(); }
  void RootOutputLariat::writeProcessHistoryRegistry() { rootOutputFile_->writeProcessHistoryRegistry(); }
  void RootOutputLariat::writeParameterSetRegistry() { rootOutputFile_->writeParameterSetRegistry(); }
  void RootOutputLariat::writeProductDescriptionRegistry() { rootOutputFile_->writeProductDescriptionRegistry(); }
  void RootOutputLariat::writeParentageRegistry() { rootOutputFile_->writeParentageRegistry(); }
  void RootOutputLariat::writeBranchIDListRegistry() { rootOutputFile_->writeBranchIDListRegistry(); }

  void
  RootOutputLariat::
  doWriteFileCatalogMetadata(FileCatalogMetadata::collection_type const & md,
                             FileCatalogMetadata::collection_type const & ssmd)
  {
    rootOutputFile_->writeFileCatalogMetadata(fstats_, md, ssmd);
  }

  void RootOutputLariat::writeProductDependencies() { rootOutputFile_->writeProductDependencies(); }
  void RootOutputLariat::finishEndFile() {
    rootOutputFile_->finishEndFile();
    fstats_.recordFileClose();
    lastClosedFileName_ =
      PostCloseFileRenamer(fstats_).
      maybeRenameFile(rootOutputFile_->currentFileName(),
                      filePattern_);
    
    // JCF, 2/26/15

    // The following code is designed to create a sentinel file of 0
    // bytes, with (almost) the same name as the freshly closed Root
    // file but with the extension ".complete" rather than ".root"


    // Make sure the name of the closed file ends in "*.root" (would
    // be shocking if it didn't, but always nice to check)

    if ( ! boost::algorithm::contains( lastClosedFileName_, ".root") ) {
      throw cet::exception("Exception in RootOutputLariat module:") <<
	" last closed file name \"" << lastClosedFileName_ << "\"" <<
	" expected to end in \".root\"!" << std::endl;
    }

    std::string sentinelFileName = lastClosedFileName_;
    boost::replace_last(sentinelFileName, ".root", ".complete");

    std::ofstream sentinelFile(sentinelFileName);
    
    if (!sentinelFile || !sentinelFile.is_open() ) {
      throw cet::exception("Exception in RootOutputLariat module:") <<
	" problem attempting to open up sentinel file \"" <<
	sentinelFileName << "\"" << std::endl;
    }
    
    sentinelFile.close();

    rootOutputFile_.reset();
  }
  bool RootOutputLariat::isFileOpen() const { return rootOutputFile_.get() != 0; }
  bool RootOutputLariat::shouldWeCloseFile() const { return rootOutputFile_->shouldWeCloseFile(); }

  void RootOutputLariat::doOpenFile() {
      if (inputFileCount_ == 0) {
        throw art::Exception(art::errors::LogicError)
          << "Attempt to open output file before input file. "
          << "Please report this to the core framework developers.\n";
      }
      rootOutputFile_.reset(new RootOutputFileLariat(this,
                                               unique_filename(tmpDir_ +
                                                               "/RootOutputLariat")));
      fstats_.recordFileOpen();
  }

  std::string const &
  RootOutputLariat::lastClosedFileName() const {
    if (lastClosedFileName_.empty()) {
      throw Exception(errors::LogicError, "RootOutputLariat::currentFileName(): ")
        << "called before meaningful.\n";
    }
    return lastClosedFileName_;
  }

}  // art

// ======================================================================

DEFINE_ART_MODULE(RootOutputLariat)

// ======================================================================
