////////////////////////////////////////////////////////////////////////
// Class:       WUTDump
// Module Type: analyzer
// File:        WUTDump_module.cc
// Description: Prints out information about each event.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "art/Utilities/Exception.h"
#include "lariat-artdaq/Overlays/WUTHitFragment.hh"
#include "ModuleUtils.hh"
#include "WUTFragment.h"
#include "artdaq-core/Data/Fragments.hh"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <vector>
#include <iostream>

namespace lariat {
  class WUTDump;
}

class lariat::WUTDump : public art::EDAnalyzer {
public:
  explicit WUTDump(fhicl::ParameterSet const & pset);
  virtual ~WUTDump() {}

  virtual void analyze(art::Event const & evt);

private:
  std::string raw_data_label_;
  uint32_t max_wuts_to_show_;
};


lariat::WUTDump::WUTDump(fhicl::ParameterSet const & pset)
    : EDAnalyzer(pset),
      raw_data_label_(pset.get<std::string>("raw_data_label")),
      max_wuts_to_show_(pset.get<uint32_t>("max_wuts_to_show", 0))
		      
{
}

void lariat::WUTDump::analyze(art::Event const & evt)
{

  art::EventNumber_t eventNumber = evt.event();

  art::Handle<artdaq::Fragments> raw;
  evt.getByLabel(raw_data_label_, "SPILL", raw);

  if ( raw.isValid() ) {

    if (raw->size() != 1 ) {
      throw cet::exception("Error in WUTDump module: expected one and only one spill fragment");
    }
    
    const auto& frag((*raw)[0]);

    std::vector<const uint8_t*> lariatFragPtrs;

    LariatFragment::LariatFragmentType lariatFragType = LariatFragment::LariatFragmentType::FRAGMENT_TYPE_WUT_HIT_HEADER;

    getLariatFragments(frag, 
     		       lariatFragType,
		       lariatFragPtrs
		       );

    std::cout << std::endl;
    std::cout << "Run " << evt.run() << ", subrun " << evt.subRun()
       	      << ", spill " << eventNumber << " has " << lariatFragPtrs.size()
   	      << " fragment(s) of type " << lariatFragType << std::endl;
      
    for (size_t idx = 0; idx < lariatFragPtrs.size(); ++idx) {

      lariat::WUTHitFragment bb( lariatFragPtrs[idx]  );

      bb.printHeader();

      if (max_wuts_to_show_ > 0) {

        uint32_t actual_to_show = max_wuts_to_show_ <= bb.total_values() ?
          max_wuts_to_show_ : bb.total_values() ;

	std::cout << std::endl;
	std::cout << "First " << actual_to_show
		  << " WUT values in the fragment: "
		  << std::endl;

	for (uint32_t i_wut = 0; i_wut < actual_to_show; ++i_wut) {
	  std::cout << "(Channel " << static_cast<int>( (*(bb.dataBegin() + i_wut)).channel ) <<
            ", Timebin " << (*(bb.dataBegin() + i_wut)).timeBin << ") ";
        }
      }

      std::cout << std::endl;
      std::cout << "######################################################################" << std::endl;
      std::cout << std::endl;
      std::cout << std::endl;
    }
  } else {
    throw cet::exception("Error in WUTDump_module: expected one and only one spill fragment");
  }

  std::cout << std::endl;

}

DEFINE_ART_MODULE(lariat::WUTDump)
