#include "Lariat.hh"
#include "LariatFragment.h"
#include "LariatException.h"
#include "CAENFragment.h"

#include "include/LariatConfiguration.h"

#include "art/Utilities/Exception.h"
#include "artdaq/Application/GeneratorMacros.hh"
#include "cetlib/exception.h"

#include "lariat-artdaq/Overlays/FragmentType.hh"
#include "fhiclcpp/ParameterSet.h"
#include "artdaq-core/Utilities/SimpleLookupPolicy.h"
#include "artdaq-core/Data/Fragment.hh"

#include <fstream>
#include <iomanip>
#include <iterator>
#include <iostream>
#include <limits>
#include <chrono>
#include <cstdlib>

#include <unistd.h>


lariat::Lariat::Lariat(fhicl::ParameterSet const & ps)
  :
  CommandableFragmentGenerator(ps),
  LariatReadout(1, ps.get<size_t>("nspills"), true),
  keyfilename_("/opt/lariat-online/keypress.txt"),
  throttle_usecs_(ps.get<size_t>("throttle_usecs", 0)),
  throttle_usecs_check_(ps.get<size_t>("throttle_usecs_check", 0)),
  seqID_(1)
  { }

// JCF, 1/9/15

// With hardware readout code now in the body of getNext_(), it's
// possible to directly check should_stop() to see if a pause or stop
// has been issued; an "alert variable" is no longer necessary

 // void lariat::Lariat::start() 
 // {
 //   std::cout << "Resetting LariatReadout PendingCommands at start()" << std::endl;
 //   PendingCommand(false);
 // }

 // void lariat::Lariat::stop() 
 // {
 //   std::cout << "Telling LariatReadout about stop()" << std::endl;
 //   PendingCommand(true);
 // }

 // void lariat::Lariat::pause() 
 // {
 //   std::cout << "Telling LariatReadout about pause()" << std::endl;
 //   PendingCommand(true);
 // }

 // void lariat::Lariat::resume()
 // {
 //   std::cout << "Resetting LariatReadout PendingCommand at resume()" 
 // 	     << std::endl;
 //   PendingCommand(false);
 // }

 // std::string lariat::Lariat::report() 
 // {
 //   std::cout << "Telling LariatReadout about report()" << std::endl;
 //   PendingCommand(true);
 //   return "";
 // }


char lariat::Lariat::CheckKeypress() {

  char key;
  size_t nread = 0;

  std::fstream keyfile( keyfilename_ );

  if ( keyfile.is_open() ) {
    keyfile.seekg( 0, keyfile.end);
    nread = keyfile.tellg();
    keyfile.seekg(0, keyfile.beg);
  }

  if (nread == 0) { // No key was pressed 
    key = '\0';
  } else if (nread == 1) {
    keyfile >> key;
    std::cerr << "Lariat generator: character \"" << key << "\" found in " << keyfilename_ << std::endl ;

  } else if (nread > 1) {
    std::cerr << "Lariat generator: keystroke file \"" << keyfilename_ << "\" is more than one character; file is ignored" << std::endl;
    key = '\0';
  }
    
  keyfile.close();

  return key;
}


bool lariat::Lariat::getNext_(artdaq::FragmentPtrs & frags) {

  std::cout << "Lariat generator: Entered getNext_()" << std::endl;

  // JCF, 9/23/14                                                                                                
  // If throttle_usecs_ is greater than zero (i.e., user requests a
  // sleep interval before generating the pseudodata) then during that
  // interval perform a periodic check to see whether a stop request
  // has been received

  // Values for throttle_usecs_ and throttle_usecs_check_ will have
  // been tested for validity in constructor

  // JCF, 1/8/15: Is this something we still want?

  if (throttle_usecs_ > 0) {

    std::cout << "Sleep time is " << throttle_usecs_ / 1000000 << " seconds" << std::endl;

    size_t nchecks = throttle_usecs_ / throttle_usecs_check_;

    for (size_t i_c = 0; i_c < nchecks; ++i_c) {
      usleep( throttle_usecs_check_ );

      if (should_stop()) {
	return false;
      }

      char key = CheckKeypress();

      if (key == 'q') {
	return false;
      }
    }
  } else {

    if (should_stop()) {
      std::cout << "Lariat generator: stop received, will return false" << std::endl;
      return false;
    }

    char key = CheckKeypress();

    if (key == 'q') {
      std::cout << "Lariat generator: q hit, will return false" << std::endl;
      return false;
    }
  }

  // JCF, 1/9/15

  // The idea behind the following two lines is that the memory
  // representing the stream of data from the hardware does NOT need
  // to be explicitly deleted via "delete" (and will cause a
  // segmentation violation if it is, crashing the program), since
  // it's wrapped with a unique_ptr; this allows for syntactic
  // consistency with the use of the output stream in the original
  // LariatReadout::readSpill() function, in which output was an
  // argument passed to readSpill, not intended for deletion within
  // the body of the funtion.
  
  std::unique_ptr<std::stringstream> output_ptr(new std::stringstream);
  std::stringstream* output = output_ptr.get();
	
  // JCF, 10/22/14

  // At Kurt's suggestion, have added a feature where if we're in
  // artdaq mode, readSpill will return after timeout_in_secs seconds
  // if no data has been taken

  // JCF, 1/9/15 -- for now, disable the feature by setting timeout to "infinity"

  double timeout_in_secs = std::numeric_limits<double>::max();

  auto lasttime = std::chrono::high_resolution_clock::now();

  // JCF, 1/9/15

  // In artdaq, other means than the "keepRunning" variable are used
  // to determine if the DAQ should complete (specifically, this
  // function returning false)

  artdaqMode = true; 
  LariatReadout::keepRunning = true; 

  struct LariatFragment::RunInfo runInfo;
  runInfo.fragmentSize         = sizeof(LariatFragment::RunInfo);
  runInfo.fragmentType         = LariatFragment::FRAGMENT_TYPE_RUNINFO;
  runInfo.runNumber            = config->runNumber;
  runInfo.runType              = config->runType;
  runInfo.recordLength         = config->v1740->recordLength;
  runInfo.nBoards              = config->v1740->nBoards;
  runInfo.link                 = config->v1740->link;
  for (uint32_t b=0; b<runInfo.nBoards; b++)
  { runInfo.boardId[b]     = config->v1740->boardId[b];}
  runInfo.postPercent          = config->v1740->postPercent;
  runInfo.maxEventsPerTransfer = config->v1740->maxEventsPerTransfer;
  runInfo.eventsPerInterrupt   = config->v1740->eventsPerInterrupt;
  runInfo.irqWaitTime          = config->v1740->irqWaitTime;
  runInfo.usePedestals         = (uint32_t)config->v1740->usePedestals;
  runInfo.dacValue             = config->v1740->dacValue;
  runInfo.samplingRate         = config->v1740->samplingRate;

  output->write((char *)&runInfo,sizeof(runInfo));
  output->write((char *)&lariatConfig,sizeof(lariatConfig.fragmentSize)+
		                      sizeof(lariatConfig.fragmentType));
  uint32_t remainder = lariatConfig.fragmentSize - (sizeof(lariatConfig.fragmentSize)+
						    sizeof(lariatConfig.fragmentType));
  output->write(lariatConfig.xml,remainder);

  auto lastsize = output->tellp();

  int errorCount =0;
  int wutPoll = 0;

  //  controller->pollReadoutRegister(1000000000,1);

  bool previousSpillActive = false;
  if ( ( controller != NULL) && ( config->v2718->spillFlag ))
  {
    previousSpillActive = controller->spill();
    //previousSpillActive = !( controller->pollReadoutRegister());
  }
  else if ( config->tdc->spillFlag && tdc != NULL )
  {
    previousSpillActive = tdc->spill();
  }
  std::cout << "SpillGate: " << previousSpillActive;
  std::cout.flush();

  // In artdaq, logic external to this function can be used to set the
  // number of desired spills (events), so just set config->events to
  // "infinity" if we're using artdaq

  if ( artdaqMode || config->events <= 0 ) { config->events = 0x7FFFFFFF;}

  for ( int e=0; spillNumber<=config->events && LariatReadout::keepRunning; e++)
  {
    bool haveV1740 = false;
    bool haveV1751 = false;
    char key = '\0';

    while ( !haveV1740 && !haveV1751 && LariatReadout::keepRunning)
    {
      // JCF, 12/29/14

      // If the size of the output stream has not increased after
      // timeout_in_secs seconds, that means we haven't gotten any
      // data from the detector; if this is the case (and we're not in
      // a spillActive state) we simply return. This feature allows
      // artdaq's "pause" and "resume" transitions to go through
      // without being blocked

      if ( artdaqMode && !spillActive) {

	auto currenttime = std::chrono::high_resolution_clock::now();

	if (std::chrono::duration_cast<std::chrono::duration<double>>(currenttime-lasttime).count() > timeout_in_secs) {
	  if (static_cast<decltype(lastsize)>(output->tellp()) == lastsize ) {
	    std::cout << "No increase in stream size in last " << timeout_in_secs << " seconds, will exit Lariat::getNext_()..." << std::endl;
	    //	    return -1;  // A negative number obviously isn't a spill number!
	    return true;
	  } else {
	    lastsize = output->tellp();
	    lasttime = currenttime;
	  }
	}
      }

      if ( v1740 != NULL ) { haveV1740 = v1740->waitForTrigger(); }

      if ( v1751 != NULL ) { haveV1751 = v1751->waitForTrigger();}

      if ( v1740 == NULL && v1751 == NULL ) { usleep(1000); }

      key = '\0';
      std::size_t nread = 0;
      if ( ! artdaqMode ) {
      	nread = read(STDIN_FILENO, &key, 1);
      } else {

	key = CheckKeypress();
	
	if (key != '\0') { 
	  nread++; 

	  std::string cmd = " : > " + keyfilename_ ;
	  int result = std::system( cmd.c_str() );

	  if (result != 0) {
	    std::cerr << "Warning in Lariat generator: nonzero return value from system() "
		      << "call designed to reset \"" << keyfilename_ << "\" to zero bytes: " 
		      << result << std::endl;
	  }
	}
      }

      if (should_stop()) {
	std::cout << "Lariat generator: \"stop\" or \"pause\" command received, returning getNext_() early"
		  << std::endl;
	LariatReadout::keepRunning = false;
      }

      // // Pending transition command?  Then return early
      // if ( PendingCommand() )
      // { 
      // 	std::cout << "PendingCommand is waiting, returning readSpill () early" 
      // 		  << std::endl;
      // 	LariatReadout::keepRunning = false;
      // }

      if (nread > 0)
      {
      	switch (key)
      	{
          case 'q': case 'Q': case 3:
      	    LariatReadout::keepRunning = false;
      	    break;
          case 'p': case 'P' :
      	    if ( v1740 != NULL )      v1740->print(); 
      	    if ( v1751 != NULL )      v1751->print(); 
      	    if ( tdc != NULL )        tdc->print();
      	    if ( wut != NULL )        wut->print();
      	    if ( trigger != NULL)     trigger->print();
	    if ( controller != NULL ) controller->print();
      	    std::cout << std::endl; 
      	    break;
          case 'h': case 'H' : case '?' :
      	    printHelp();
      	    break;
      	}
      }

      if ( controller != NULL && config->v2718->spillFlag )
      {
	spillActive = controller->spill();
	//spillActive = !( controller->pollReadoutRegister());
	//	std::cout << spillActive ;  std::cout.flush();
      }
      else if ( tdc != NULL && config->tdc->spillFlag > 0 )
      {
	spillActive = tdc->spill();
      }

      if ( haveV1740 )
      {
	try
	{
	  v1740->readEvent(output);
	}
	catch ( LariatException e )
	{
	  std::cout << std::endl;
	  e.print();
	  errorCount++;
	  bool keepTrying = true;
	  while ( keepTrying )
	  {
	    try
	    {
	      lariatReset();
	      configure();
	      keepTrying = false;
	    }
	    catch ( LariatException e ) { e.print();}
	  }
	}
	countV1740++;
	spillV1740++;

	// JCF, 6/16/14

	// The fact that every message in artdaq has a date prepended
	// on screen means that we want to be a little more sparing in
	// our output than if this function was called as part of the
	// LariatRun executable

	if (!artdaqMode || (artdaqMode && countV1740 % 30 == 0) ) 
	{
	  std::cout << spillActive;
	}
	std::cout.flush();
      }

      if ( haveV1751 )
      {
	try
	{
	  v1751->readEvent(output);
	}
	catch ( LariatException e )
	{
	  std::cout << std::endl;
	  e.print();
	  errorCount++;
	  bool keepTrying = true;
	  while ( keepTrying )
	  {
	    try
	    {
	      lariatReset();
	      configure();
	      keepTrying = false;
	    }
	    catch ( LariatException e ) { e.print();}
	  }
	}
	countV1751++;
	spillV1751++;

	if (!artdaqMode || (artdaqMode && countV1751 % 30 == 0) ) 
	{
	  std::cout << spillActive ;
	  std::cout.flush();
	}
      }

      wutPoll++;
      int wutInterval = 0;
      if ( config->wut->readoutPeriod > 0 )
      { wutInterval = wutPoll % config->wut->readoutPeriod;}

      if ( (wutInterval == 0 ) && ( wut!=NULL) )
      {
	wut->readEvent(output,wutPoll);
	std::cout.flush();
	printNusoft();
      }


      // Now we need to check if the end-of-spill has occurred 
      // We have two possible sources of the spill flag:  the Wire 
      // Chamber TDC controller or the VME V2718 crate controller, 
      // so this is a skosh complicated.   Preferentially we will 
      // use the V2718, since this is the normal mode of Lariat 
      // operation.

      // The spillIsDone flag should only be true if we 
      // observe a TRANSITION from "spill on" to "spill off"
      bool spillIsDone = false;

      // Is the V2718 controller present?   If yes, use it!
      if ( ( controller != NULL) && ( config->v2718->spillFlag ))
      {
	//spillActive = !( controller->pollReadoutRegister());
        spillActive = controller->spill();
      }
      else if ( ( tdc != NULL ) && ( config->tdc->spillFlag ) )
      {
	spillActive = tdc->spill();
      }

      if ( spillActive != previousSpillActive )
      {
	if ( previousSpillActive ) 
	{ 
	  spillIsDone = true;
	}
	previousSpillActive = spillActive;
      }

      // If we have really found the end of spill, then we need to 
      // read out trigger, WUT, W/C TDCs, close the data file and 
      // write out the web status files:

      if(spillIsDone) 
      {
	if ( trigger != NULL )
	{
	  trigger->readEvent(output);
	  // trigger.getFragment().print();
	  
	  std::ofstream of0("/nusoft/app/web/htdoc/lariat/channels.html");
	  std::ofstream of1("/nusoft/app/web/htdoc/lariat/triggers.html");
	  std::ofstream of2("/nusoft/app/web/htdoc/lariat/vetos.html");
	  trigger->getFragment().printHTML(0, of0);
	  trigger->getFragment().printHTML(1, of1);
	  trigger->getFragment().printHTML(2, of2);
	  printNusoft();
	}

	// we rely on counters being reset prior to each beam spill by
	// a hardware input
	if ( controller != NULL )
	{
	  controller->resetReadoutRegister();
	}

	// The WUT is asynchronous and double buffered, so we need to 
	// read out any straggler hits left from after the last time 
	// we read it:
	if ( wut != NULL ) 
	{ 
	  wut->readEvent(output,wutPoll);
	  wutPoll++;
	}

	// Could be left-over triggers in the V1740s and V1751
	if ( v1740 != NULL )
	{
	  while ( v1740->readEvent(output) > 0 ) 
	  {
	    countV1740++;
	    spillV1740++;
	  }
	}

	if ( v1751 != NULL )
	{
	  while ( v1751->readEvent(output) > 0 ) 
	  {
	    countV1751++;
	    spillV1751++;
	  }
	}

	// The W/C TDC system is designed to cache all hits during 
	// a spill in memory, and only read out when the spill 
	// is over, so we read the whole dang thing now:
	if ( ( tdc != NULL ) && 
	     config->tdc->enableReadout &&
	     ( config->tdc->nBoards > 0 ) )
	{
	  tdc->readEvent(output);
	}

	std::cout << std::endl << "Spill " << spillNumber
		  << " has ended" << std::endl;
	std::cout << std::endl << "Read " << spillV1740 << "/" <<
	  countV1740 <<  " V1740 triggers and "
		  << spillV1751 << "/" << 
	  countV1751 << " V1751 triggers" << std::endl;

	writeSpillTrailer(output);

	spillV1740 = 0;
	spillV1751 = 0;

	spillNumber++;
	wutPoll = 0;

	if ( spillNumber > config->events )
	{ 
	  LariatReadout::keepRunning = false;
	}
	std::cout.flush();

	std::unique_ptr<artdaq::Fragment> frag( new artdaq::Fragment);

	output->seekg(0);

	frag->resize( ceil( output->str().size() / static_cast<double>( sizeof(artdaq::RawDataType ) )));
	output->read( reinterpret_cast<char*>( &*(frag->dataBegin()) ), output->str().size() );

	frag->setSequenceID( seqID_++ );
	frag->setFragmentID( 0 );
	frag->setUserType( FragmentType::SPILL );  

	frags.emplace_back( std::move( frag ) );

	std::cout << "Lariat generator: standard close to getNext_(), will return true" << std::endl;

	return true;
      }
    }
  }

  if ( !artdaqMode && ( spillNumber > config->events ) )
  { LariatReadout::keepRunning = false;}

  std::cout << std::endl << "Read " << countV1740 << " V1740 triggers and "
	    << countV1751 << " V1751 triggers" << std::endl;
  return false;
}

// The following macro is defined in artdaq's GeneratorMacros.hh header
DEFINE_ARTDAQ_COMMANDABLE_GENERATOR(lariat::Lariat) 
