#include "SpillFileReader.hh"

#include "LariatFragment.h"
#include "CAENFragment.h"

#include "art/Utilities/Exception.h"
#include "artdaq/Application/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "lariat-artdaq/Overlays/FragmentType.hh"
#include "fhiclcpp/ParameterSet.h"
#include "artdaq-core/Utilities/SimpleLookupPolicy.h"

#include <fstream>
#include <iomanip>
#include <iterator>
#include <iostream>
#include <limits>

#include <unistd.h>


lariat::SpillFileReader::SpillFileReader(fhicl::ParameterSet const & ps)
  :
  CommandableFragmentGenerator(ps),
  input_filelist_( ps.get<std::string>("input_filelist") ),
  throttle_usecs_(ps.get<size_t>("throttle_usecs", 100000)),
  throttle_usecs_check_(ps.get<size_t>("throttle_usecs_check", 10000)),
  fragment_ids_{ ps.get< std::vector<artdaq::Fragment::fragment_id_t> >("fragment_ids") },
  infilelist_( input_filelist_.c_str() )//,
  //  output_directory_( ps.get<std::string>("output_directory") )
  {

    if ( ! infilelist_.is_open() ) {
      throw cet::exception("SpillFileReader: Unable to open input filelist");
    }

    if (throttle_usecs_ > 0 && (throttle_usecs_check_ >= throttle_usecs_ ||
				throttle_usecs_ % throttle_usecs_check_ != 0) ) {
      throw cet::exception("Error in SpillFileReader: disallowed combination of throttle_usecs and throttle_usecs_check (see SpillFileReader.hh for rules)");
    }
  }


bool lariat::SpillFileReader::getNext_(artdaq::FragmentPtrs & frags) {

  static artdaq::Fragment::sequence_id_t seqID = 1;
  std::string current_filename;

  // JCF, 9/23/14                                                                                                
  // If throttle_usecs_ is greater than zero (i.e., user requests a
  // sleep interval before generating the pseudodata) then during that
  // interval perform a periodic check to see whether a stop request
  // has been received

  // Values for throttle_usecs_ and throttle_usecs_check_ will have
  // been tested for validity in constructor

  char key;
  size_t nread = 0;

  std::string keyfilename = "/opt/lariat-online/keypress.txt";

  if (throttle_usecs_ > 0) {
    size_t nchecks = throttle_usecs_ / throttle_usecs_check_;

    for (size_t i_c = 0; i_c < nchecks; ++i_c) {
      usleep( throttle_usecs_check_ );

      if (should_stop()) {
	return false;
      }

      // Now check to see if the "q" button was hit

      std::fstream keyfile( keyfilename );

      if ( keyfile.is_open() ) {
	keyfile.seekg( 0, keyfile.end);
	nread = keyfile.tellg();
	keyfile.seekg(0, keyfile.beg);
      }

      if (nread == 0) { // No key was pressed 
	continue; 
      } else if (nread == 1) {
	keyfile >> key;
	std::cerr << "LariatReadout: character \"" << key << "\" found in " << keyfilename << std::endl ;

	keyfile.close();

	// Reset the keystroke file to 0 bytes

	std::string cmd = " : > " + keyfilename;
	int result = system( cmd.c_str() );

	if (result != 0) {
	  std::cerr << "Warning in LariatReadout: nonzero return value from system() call designed to reset \"" << keyfilename << "\" to zero bytes" << std::endl;
	}

	if (key != 'q') {
	  throw cet::exception("Keypress should only be 'q' ");
	}

	
	size_t nread2 = 0;
	std::fstream keyfile2( keyfilename );
	keyfile2.seekg( 0, keyfile2.end);
	nread2 = keyfile2.tellg();
	keyfile2.seekg(0, keyfile2.beg);

	std::cout << keyfilename << " should be empty after running \"" << cmd << "\"; found " << nread2 << " bytes" << std::endl;

	return false;
      } else if (nread > 1) {
	std::cerr << "Warning in LariatReadout: keystroke file \"" << keyfilename << "\" is more than one character" << std::endl;
      }
    }
  } else {
    if (should_stop()) {
      return false;
    }
  }


  if ( ! std::getline( infilelist_, current_filename) ) {
    std::cout << "No more files listed in " << input_filelist_ << std::endl;
    return false;
  }

  std::cout << "Opening " << current_filename << std::endl;

  if (infile_.is_open()) {
    infile_.close();
  }

  infile_.open( current_filename.c_str() , std::ios::binary );
    
  if ( !infile_.is_open() ) {
    std::string errorMessage = "Error in SpillFileReader: Couldn't open input file " + current_filename ;
    throw cet::exception(errorMessage);
  }

  std::unique_ptr<artdaq::Fragment> frag( new artdaq::Fragment);

  infile_.seekg(0, std::ifstream::end);
  std::ifstream::pos_type filesize = infile_.tellg();
  infile_.seekg(0);

  std::cout << "Input file is " << filesize << " bytes" << std::endl;

  frag->resize( ceil( filesize / static_cast<double>(sizeof(artdaq::RawDataType)) ) );

  infile_.read( reinterpret_cast<char*>( &*(frag->dataBegin() )), filesize);

  frag->setFragmentID( 1  );
  frag->setSequenceID( seqID );
  frag->setUserType( lariat::detail::FragmentType::SPILL );

  seqID++;

  frags.emplace_back( std::move( frag ) );

  return true;
}

// The following macro is defined in artdaq's GeneratorMacros.hh header
DEFINE_ARTDAQ_COMMANDABLE_GENERATOR(lariat::SpillFileReader) 
