//
// CAENChain.cc    (W.Badgett)
//  
// Class to handle the configuration and readout of a chain 
// of CAEN CAEN wave form digitizers
//
// NB: Mixed chains not currently implemented, only pure V1740 
//     or V1751 chains allowed for now
//
#include <iostream>
#include "CAENChain.h"
#include "CAENDriver.h"
#include "LariatConfiguration.h"
#include "LariatException.h"

CAENChain::CAENChain(LariatConfiguration *config_, 
		     CAENConfiguration   *caen_,
		     int chainType_)
{
  config    = config_;
  caen      = caen_;
  chainType = chainType_;

  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    {
      if ( caen->enableBoard[b] )
      {
	v1740[b] = new V1740Driver(config,caen,b,caen->boardId[b]);
	if ( v1740[b]->ok )
	{
	  std::cout << "V1740 Board " << b << " BoardId " << caen->boardId[b] <<
	    " opened OK"<<std::endl;
	}
	else
	{
	  std::cout << "V1740 Board " << b << " BoardId " <<  caen->boardId[b] <<
	    " failed to open" << std::endl;
	}
      }
      else { v1740[b] = NULL;}
    }
  }

  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    {
      if ( caen->enableBoard[b] )
      {
	v1751[b] = new V1751Driver(config,caen,b,caen->boardId[b]);
	if ( v1751[b]->ok )
	{ std::cout << "V1751 Board " << b << " opened OK, Id "<<
	    caen->boardId[b] << std::endl;}
	else
	{ std::cout << "V1751 Board " << b << " failed to open" << std::endl;}
      }
      else { v1751[b] = NULL;}
    }
  }
}

CAENChain::~CAENChain()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) { delete(v1740[b]);} }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) delete(v1751[b]); }
  }
}

uint32_t CAENChain::readEvent(std::ostream *dataFile)
{
  uint32_t nEvents[MAX_BOARDS];

  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) nEvents[b] = v1740[b]->readEvent(dataFile); }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) nEvents[b] = v1751[b]->readEvent(dataFile); }
  }

  int firstEvents  = -1;
  int firstBoardId = -1;
  for ( int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
  {
    if ( firstEvents < 0 )
    {
      if ( caen->enableBoard[b] ) 
      { 
	firstEvents  = nEvents[b];
	firstBoardId = caen->boardId[b];
      }
    }
    else
    {
      if ( ( firstEvents >= 0 ) && ( caen->enableBoard[b] ))
      {
	if ( ( firstEvents != (int)nEvents[b] ) && ( nEvents[b] > 1 ))
	{
	  char text[256];
	  sprintf(text,"BoardId %d eventCounter at %d does not match boardId %d at %d", 
		  caen->boardId[b], nEvents[b], firstBoardId, firstEvents);
	  std::string label(text);
	  LariatException e(CAEN_DGTZ_BadEventNumber,label,caen->boardId[b]);
	  std::cout << std::endl << text << std::endl;
	  throw(e);
	}
      }
    }
  }
  return(firstEvents);
}

bool CAENChain::waitForTrigger()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    if ( CAENConfiguration::MAX_BOARDS > 0 )
    { return(v1740[0]->waitForTrigger()); }
    else { return(false);}
  }
  else if ( chainType == V1751_CHAIN )
  {
    if ( CAENConfiguration::MAX_BOARDS > 0 )
    { return(v1751[0]->waitForTrigger()); }
    else { return(false);}
  }
  else { return(false);}
}

void CAENChain::fireSWTrigger()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1740[b]->fireSWTrigger(); }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1751[b]->fireSWTrigger(); }
  }
}

void CAENChain::configure()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1740[b]->configure(); }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1751[b]->configure(); }
  }
}

void CAENChain::configurePedestal()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1740[b]->configurePedestal(); }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1751[b]->configurePedestal(); }
  }
}

void CAENChain::configurePhysics()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) { v1740[b]->configurePhysics(); } }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1751[b]->configurePhysics(); }
  }
}

void CAENChain::writePedestals(std::ostream * pedestalFile)
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1740[b]->writePedestals(pedestalFile); }
  }
  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    { if ( caen->enableBoard[b] ) v1751[b]->writePedestals(pedestalFile); }
  }
}

void CAENChain::print()
{
  if ( chainType == V1740_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    {
      if ( caen->enableBoard[b] )
      {
	v1740[b]->readConfiguration();
	v1740[b]->print();
      }
    }
  }

  if ( chainType == V1751_CHAIN || chainType == MIXED_CHAIN )
  {
    for (int b=0; b<CAENConfiguration::MAX_BOARDS; b++)
    {
      if ( caen->enableBoard[b] )
      {
	v1751[b]->readConfiguration();
	v1751[b]->print();
      }
    }
  }
}
