//
//  CAENDriver.cc  (W.Badgett)
//
//  Base class to configure and readout one 
//   CAEN digitizer
//

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <ptree.hpp>
#include "CAENDriver.h"
#include "CAENFragment.h"
#include "LariatConfiguration.h"
#include "LariatFragment.h"
#include "LariatReadout.h"
#include "CAENDecoder.h"
#include "LariatException.h"

CAENDriver::~CAENDriver()
{
  CAEN_DGTZ_ErrorCode retcod;

  if ( handle >= 0 )
  {
    // Make sure DAQ is off
    retcod = CAEN_DGTZ_SetAcquisitionMode(handle,CAEN_DGTZ_SW_CONTROLLED);
    CAENDecoder::checkError(retcod,"SetAcquisitionMode",boardId);
    
    retcod = CAEN_DGTZ_SWStopAcquisition(handle);
    CAENDecoder::checkError(retcod,"SWStartAcquisition",boardId);

    retcod = CAEN_DGTZ_FreeReadoutBuffer(&buffer);
    CAENDecoder::checkError(retcod,"FreeReadoutBuffer",boardId);

    retcod = CAEN_DGTZ_FreeEvent(handle,(void **)&event);
    CAENDecoder::checkError(retcod,"FreeEvent",boardId);
    
    retcod = CAEN_DGTZ_CloseDigitizer(handle);
    CAENDecoder::checkError(retcod,"CloseDigitizer",boardId);
  }
}


void CAENDriver::setRecordFormat()
{
  CAEN_DGTZ_ErrorCode retcod;

  retcod = CAEN_DGTZ_SetRecordLength(handle,caen->recordLength);
  CAENDecoder::checkError(retcod,"SetRecordLength",boardId);

  recordLength = 0;
  retcod = CAEN_DGTZ_GetRecordLength(handle,&recordLength);
  CAENDecoder::checkError(retcod,"GetRecordLength",boardId);
  if ( (int)recordLength != caen->recordLength )
  {
    char text[128];
    sprintf(text,"SetRecordLength wrote %d read %d", 
	    caen->recordLength, recordLength);
    LariatException e(CAEN_DGTZ_DigitizerNotReady,
		      "SetRecordLength", boardId);
    e.print();
    throw(e);
  }

  retcod = CAEN_DGTZ_SetPostTriggerSize(handle,caen->postPercent);
  CAENDecoder::checkError(retcod,"SetPostTriggerSize",boardId);
  
  retcod = CAEN_DGTZ_GetPostTriggerSize(handle,&postPercent);
  CAENDecoder::checkError(retcod,"GetPostTriggerSize",boardId);
  if ( (int)postPercent != caen->postPercent )
  {
    char text[128];
    sprintf(text,"SetPostPercent wrote %d read %d", 
	    caen->postPercent, postPercent);
    LariatException e(CAEN_DGTZ_DigitizerNotReady,text, boardId);
    e.print();
    throw(e);
  }
}

void CAENDriver::configure()
{
  CAEN_DGTZ_ErrorCode retcod;

  retcod =  CAEN_DGTZ_Reset(handle);
  CAENDecoder::checkError(retcod,"Reset",boardId);
  usleep(100000);

  // Make sure DAQ is off first
  retcod = CAEN_DGTZ_SetAcquisitionMode(handle,CAEN_DGTZ_SW_CONTROLLED);
  CAENDecoder::checkError(retcod,"SetAcquisitionMode",boardId);
    
  retcod = CAEN_DGTZ_SWStopAcquisition(handle);
  CAENDecoder::checkError(retcod,"SWStartAcquisition",boardId);

  retcod =  CAEN_DGTZ_GetInfo(handle, &boardInfo);
  ok = ( retcod == CAEN_DGTZ_Success );

  // Must be done prior to MallocReadoutBuffer
  setRecordFormat();

  buffer = NULL;
  size = 0;
  retcod = CAEN_DGTZ_MallocReadoutBuffer(handle,&buffer,&size);
  CAENDecoder::checkError(retcod,"MallocReadoutBuffer",boardId);

  // Allocate an event descriptor
  retcod = CAEN_DGTZ_AllocateEvent(handle,(void **)&event);
  CAENDecoder::checkError(retcod,"AllocateEvent",boardId);

  swTriggerMode = (CAEN_DGTZ_TriggerMode_t)caen->swTrgMode;
  retcod = CAEN_DGTZ_SetSWTriggerMode(handle,swTriggerMode);
  CAENDecoder::checkError(retcod,"SetSWTriggerMode",boardId);
  
  extTriggerMode = (CAEN_DGTZ_TriggerMode_t)caen->extTrgMode;
  retcod = CAEN_DGTZ_SetExtTriggerInputMode(handle,extTriggerMode);
  CAENDecoder::checkError(retcod,"SetExtTriggerInputMode",boardId);
  
  for (uint32_t c = 0; c<nChannels; c++)
  { 
    triggerPolarity = (CAEN_DGTZ_TriggerPolarity_t)caen->triggerPolarity;
    retcod = CAEN_DGTZ_SetTriggerPolarity(handle,c,triggerPolarity);
    CAENDecoder::checkError(retcod,"SetTriggerPolarity",boardId);
  }
}

CAENDriver::CAENDriver(LariatConfiguration * config_,
		       CAENConfiguration * caen_,
		       uint32_t boardChainNumber_,
		       uint32_t boardId_):
  handle(-1),
  boardId(-1),
  boardInfo(),
  ok(false),
  state(CAEN_DGTZ_ENABLE),
  mode(CAEN_DGTZ_IRQ_MODE_ROAK),
  triggerPolarity(CAEN_DGTZ_TriggerOnRisingEdge),
  swTriggerMode(CAEN_DGTZ_TRGMODE_DISABLED),
  extTriggerMode(CAEN_DGTZ_TRGMODE_DISABLED),
  zeroMode(CAEN_DGTZ_ZS_NO),
  analogMode(CAEN_DGTZ_AM_TEST),
  runSyncMode(CAEN_DGTZ_RUN_SYNC_Disabled),
  outputSignalMode(CAEN_DGTZ_TRIGGER),
  ioLevel(CAEN_DGTZ_IOLevel_NIM),
  acqMode(CAEN_DGTZ_SW_CONTROLLED),
  interruptLevel(0),
  eventNumber(0),
  statusId(0),
  numEvents(0),
  recordLength(0),
  postPercent(0),
  enableMask(0),
  nChannels(0),
  acquisitionStatus(0),
  boardReady(0),
  pllStatus(0),
  pllBypass(0),
  clockSource(0),
  eventFull(0),
  eventReady(0),
  runEnabled(0),
  buffer(NULL),
  event(NULL)
{
  CAEN_DGTZ_ErrorCode retcod;
  
  config           = config_;
  caen             = caen_;
  boardChainNumber = boardChainNumber_;
  boardId          = boardId_;

  // GetBoardInfo returns the incorrect number of channels
  nChannels        = caen->nChannels;

  retcod = CAEN_DGTZ_OpenDigitizer(CAEN_DGTZ_PCI_OpticalLink,
   				   caen->link,
   				   boardChainNumber,
   				   0,
   				   &handle);
  if ( retcod == CAEN_DGTZ_Success )
  {
    configure();
  }
  else
  {
    std::cout << "*** Fatal error opening CAEN board at " << 
      caen->link << "," << boardChainNumber << std::endl;
    CAENDecoder::checkError(retcod,"OpenDigitizer",boardId);
    handle = -1;
  }
}

void CAENDriver::configureCommon()
{
  CAEN_DGTZ_ErrorCode retcod;

  numEvents   = caen->maxEventsPerTransfer;
  retcod = CAEN_DGTZ_SetMaxNumEventsBLT(handle,numEvents);
  CAENDecoder::checkError(retcod,"SetMaxNumEventsBLT",boardId);

  runSyncMode = (CAEN_DGTZ_RunSyncMode_t)caen->runSyncMode;
  retcod = CAEN_DGTZ_SetRunSynchronizationMode(handle,runSyncMode);
  CAENDecoder::checkError(retcod,"SetRunSynchronizationMode",boardId);

  outputSignalMode = (CAEN_DGTZ_OutputSignalMode_t)caen->outputSignalMode;
  // Following set function not yet implemented (!)
  //  retcod = CAEN_DGTZ_SetOutputSignalMode(handle,outputSignalMode);
  // CAENDecoder::checkError(retcod,"SetOutputSignalMode");

  // For completeness; we don't use this output
  analogMode = CAEN_DGTZ_AM_TRIGGER_MAJORITY; 
  retcod = CAEN_DGTZ_SetAnalogMonOutput(handle,analogMode);
  CAENDecoder::checkError(retcod,"SetAnalogMonOutput",boardId);
}

void CAENDriver::configurePedestal()
{
  CAEN_DGTZ_ErrorCode retcod;
  configureCommon();

  // Turn off interrupts
  state           = CAEN_DGTZ_DISABLE;
  interruptLevel  = 1; 
  statusId        = caen->boardId[boardChainNumber];
  eventNumber     = 1; 
  mode            = CAEN_DGTZ_IRQ_MODE_ROAK; // Default mode

  if ( boardChainNumber == 0 )
  {
    state       = CAEN_DGTZ_ENABLE;
    std::cout << "Enabling IRQs " << boardChainNumber << " " 
	      << boardId << std::endl;
  }
  retcod = CAEN_DGTZ_SetInterruptConfig(handle,state,interruptLevel,
  					statusId,eventNumber,mode);
  CAENDecoder::checkError(retcod,"SetInterruptConfig",boardId);

  swTriggerMode = CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT;
  retcod = CAEN_DGTZ_SetSWTriggerMode(handle,swTriggerMode);
  CAENDecoder::checkError(retcod,"SetSWTriggerMode",boardId);

  extTriggerMode = CAEN_DGTZ_TRGMODE_DISABLED;
  retcod = CAEN_DGTZ_SetExtTriggerInputMode(handle,extTriggerMode);
  CAENDecoder::checkError(retcod,"SetExtTriggerInputMode",boardId);

  acqMode = CAEN_DGTZ_SW_CONTROLLED;
  retcod = CAEN_DGTZ_SetAcquisitionMode(handle,acqMode);
  CAENDecoder::checkError(retcod,"SetAcquisitionMode",boardId);

  retcod = CAEN_DGTZ_SWStartAcquisition(handle);
  CAENDecoder::checkError(retcod,"SWStartAcquisition",boardId);

  // Zero sums
  for ( int c=0; c<MAX_UINT16_CHANNEL_SIZE; c++)
  {
    sum[c] = 0.0;
    sumsq[c] = 0.0;
    nSamples[c] = 0;
  }
}

void CAENDriver::configurePhysics()
{
  CAEN_DGTZ_ErrorCode retcod;

  configureCommon();

  // Turn ON interrupts for first board
  interruptLevel = 1; // Default level
  statusId       = 0;
  eventNumber    = 1; // Number of events to cause an IRQ
  mode           = CAEN_DGTZ_IRQ_MODE_ROAK; // Default mode
  state          = CAEN_DGTZ_DISABLE;
  if ( boardChainNumber == 0 )
  {
    state       = CAEN_DGTZ_ENABLE;
    std::cout << "Enabling IRQs " << boardChainNumber << " " 
	      << boardId << std::endl;
  }
  retcod = CAEN_DGTZ_SetInterruptConfig(handle,state,interruptLevel,
					statusId,eventNumber,mode);
  CAENDecoder::checkError(retcod,"SetInterruptConfig",boardId);

  swTriggerMode = (CAEN_DGTZ_TriggerMode_t)caen->swTrgMode;
  retcod = CAEN_DGTZ_SetSWTriggerMode(handle,swTriggerMode);
  CAENDecoder::checkError(retcod,"SetSWTriggerMode",boardId);

  extTriggerMode = (CAEN_DGTZ_TriggerMode_t)caen->extTrgMode;
  retcod = CAEN_DGTZ_SetExtTriggerInputMode(handle,extTriggerMode);
  CAENDecoder::checkError(retcod,"SetExtTriggerInputMode",boardId);

  acqMode = (CAEN_DGTZ_AcqMode_t)caen->acqMode;
  retcod = CAEN_DGTZ_SetAcquisitionMode(handle,acqMode);
  CAENDecoder::checkError(retcod,"SetAcquisitionMode",boardId);

  CAEN_DGTZ_AcqMode_t readBack;
  retcod = CAEN_DGTZ_GetAcquisitionMode(handle,&readBack);
  CAENDecoder::checkError(retcod,"GetAcquisitionMode",boardId);
  if ( acqMode != readBack )
  {
    char text[128];
    sprintf(text,"SetAcquisitionMode wrote %d read %d", 
	    acqMode, readBack);
    LariatException e(CAEN_DGTZ_DigitizerNotReady,text,boardId);
    e.print();
    throw(e);

  }

  retcod = CAEN_DGTZ_SWStartAcquisition(handle);
  CAENDecoder::checkError(retcod,"SWStartAcquisition",boardId);
}

void CAENDriver::readConfiguration()
{
  //CAEN_DGTZ_ErrorCode retcod;
  CAEN_DGTZ_GetInterruptConfig(handle,&state,&interruptLevel,
  				&statusId,&eventNumber,&mode);
  CAEN_DGTZ_GetMaxNumEventsBLT(handle,&numEvents);
  CAEN_DGTZ_GetSWTriggerMode(handle,&swTriggerMode);
  CAEN_DGTZ_GetExtTriggerInputMode(handle,&extTriggerMode);
  CAEN_DGTZ_GetRunSynchronizationMode(handle,&runSyncMode);
  CAEN_DGTZ_GetIOLevel(handle,&ioLevel);
  CAEN_DGTZ_GetOutputSignalMode(handle,&outputSignalMode);
  CAEN_DGTZ_GetGroupEnableMask(handle,&enableMask);
  CAEN_DGTZ_GetRecordLength(handle,&recordLength);
  CAEN_DGTZ_GetPostTriggerSize(handle,&postPercent);
  CAEN_DGTZ_GetAcquisitionMode(handle,&acqMode);
  CAEN_DGTZ_GetAnalogMonOutput(handle,&analogMode);
  readAcquisitionStatus();
}


void CAENDriver::print(std::ostream & os)
{
  os << "\nConnected to CAEN Digitizer Model " <<
    boardInfo.ModelName << " Number " << boardChainNumber << ", recognized as board type " <<
	 boardInfo.Model<< std::endl;
  os << "\tChannels        " <<  boardInfo.Channels<< std::endl;
  os << "\tFormFactor      " <<  boardInfo.FormFactor<< std::endl;
  os << "\tFamilyCode      " <<  boardInfo.FamilyCode<< std::endl;
  os << "\tSerialNumber    " <<  boardInfo.SerialNumber<< std::endl;
  os << "\tADC_NBits       " <<  boardInfo.ADC_NBits<< std::endl;
  os << "\tPCB_Revision    " <<  boardInfo.PCB_Revision<< std::endl;
  os << "\tROC FPGA Release is " <<  boardInfo.ROC_FirmwareRel<< std::endl;
  os << "\tAMC FPGA Release is " <<  boardInfo.AMC_FirmwareRel<< std::endl;
  os << "Interrupt Configuration"<< std::endl;
  os << "\tEnableDisable   " <<  state<< std::endl;
  os << "\tInterruptLevel  " <<  (int)interruptLevel<< std::endl;
  os << "\tStatusId        " <<  statusId<< std::endl;
  os << "\tEventNumber     " <<  eventNumber<< std::endl;
  os << "\tIRQMode         " << mode << " " << CAENDecoder::IRQMode(mode)<< std::endl;
  os << "Readout Configuration"<< std::endl;
  os << "\tMaxNumEvents    " <<  numEvents<< std::endl;
  os << "\tnChannels       " <<  nChannels<< std::endl;
  os << "Trigger Configuration\n"<< std::endl;
  os << "\tSWTriggerMode        " << swTriggerMode << " " << 
	 CAENDecoder::TriggerMode(swTriggerMode) << std::endl;
  os << "\tExtTriggerMode       " <<  extTriggerMode << " " <<
	 CAENDecoder::TriggerMode(extTriggerMode) << std::endl;
  os << "\tTriggerPolarity      " <<  triggerPolarity << " " <<
	 CAENDecoder::TriggerPolarity(triggerPolarity)<< std::endl;
  os << "\tRunSyncMode          " <<  runSyncMode << " " <<
	 CAENDecoder::RunSynchronizationMode(runSyncMode)<< std::endl;
  os << "\tIOLevel              " <<  ioLevel << " " << 
	 CAENDecoder::IOLevel(ioLevel)<< std::endl;
  os << "\tOutputSignalMode     " <<  outputSignalMode << " " <<
	 CAENDecoder::OutputSignalMode(outputSignalMode)<< std::endl;
  os << "Acquisition Configuration"<< std::endl;
  os << "\tGroupEnable     0x" << std::hex << enableMask << std::dec << std::endl;
  os << "\tRecordLength    " <<  recordLength<< std::endl;
  os << "\tPostTrigger     " <<  postPercent<< "%" <<std::endl;
  os << "\tAcquisitionMode " <<  acqMode << " " << 
    CAENDecoder::AcquisitionMode(acqMode)<< std::endl;
  os << "\tAnalogMonOutput " <<  analogMode << " " <<
    CAENDecoder::AnalogMonOutput(analogMode)<< std::endl;
  os << "Acquisition Status\n"<< std::endl;
  os << "  BoardReady " << boardReady <<  " PLLStatus  " <<
    pllStatus << "  PLLBypass  " << pllBypass << " ClockSource "<< 
    clockSource << std::endl;
  os << "  EventFull  " << eventFull << " EventReady " << 
    eventReady << " " << " RunEnabled " << runEnabled << std::endl;
}


void CAENDriver::fireSWTrigger()
{
  CAEN_DGTZ_ErrorCode retcod;

  retcod = CAEN_DGTZ_SendSWtrigger(handle);
  CAENDecoder::checkError(retcod,"SendSWtrigger",boardId);
}

bool CAENDriver::waitForTrigger()
{
  CAEN_DGTZ_ErrorCode retcod;

  retcod = CAEN_DGTZ_RearmInterrupt(handle);
  CAENDecoder::checkError(retcod,"RearmInterrupt",boardId);

  retcod = CAEN_DGTZ_IRQWait(handle,caen->irqWaitTime);
  if ( retcod != CAEN_DGTZ_Timeout ) 
  { CAENDecoder::checkError(retcod,"IRQWait",boardId);}

  readAcquisitionStatus();

  //  if ( eventReady ){  print();}
  //  return(retcod == CAEN_DGTZ_Success);
  return(eventReady);
}

void CAENDriver::readAcquisitionStatus()
{
  CAEN_DGTZ_ErrorCode retcod;
  retcod = CAEN_DGTZ_ReadRegister(handle,ACQUISITION_STATUS_ADDR,
				  &acquisitionStatus);
  CAENDecoder::checkError(retcod,"ReadRegisgter ACQUISITION_STATUS",boardId);

  boardReady  = (acquisitionStatus >> BOARD_READY_S )  & 1;
  pllStatus   = (acquisitionStatus >> PLL_STATUS_S )   & 1;
  pllBypass   = (acquisitionStatus >> PLL_BYPASS_S )   & 1;
  clockSource = (acquisitionStatus >> CLOCK_SOURCE_S ) & 1;
  eventFull   = (acquisitionStatus >> EVENT_FULL_S )   & 1;
  eventReady  = (acquisitionStatus >> EVENT_READY_S )  & 1;
  runEnabled  = (acquisitionStatus >> RUN_ENABLED_S )  & 1;
}


uint32_t CAENDriver::readEvent(std::ostream *dataFile)
{
  CAEN_DGTZ_ErrorCode retcod;
  uint32_t bsize = 0;
  struct timespec zeit;
  uint32_t eventCounter = 0;
  uint32_t nEvents = 0;

  retcod  = CAEN_DGTZ_ReadData(handle,CAEN_DGTZ_SLAVE_TERMINATED_READOUT_MBLT,
			       buffer,&bsize);
  if ( bsize == 0 ) {return(0);}

  CAENDecoder::checkError(retcod,"ReadData",boardId);

  retcod = CAEN_DGTZ_GetNumEvents(handle,buffer,bsize,(uint32_t *)&nEvents);
  CAENDecoder::checkError(retcod,"GetNumEvents",boardId);

  for ( uint32_t e=0; e<nEvents; e++)
  {
    clock_gettime(CLOCK_REALTIME,&zeit);

    char * evtptr = NULL;
    CAEN_DGTZ_EventInfo_t eventInfo;
    struct CAENFragment::CaenHeader header;

    retcod = CAEN_DGTZ_GetEventInfo(handle,buffer,bsize,e,&eventInfo,&evtptr);
    CAENDecoder::checkError(retcod,"GetEventInfo",boardId);

    header.fragmentType   = LariatFragment::FRAGMENT_TYPE_CAEN_ADC;
    header.eventCounter   = eventInfo.EventCounter;
    eventCounter          = header.eventCounter;
    header.eventSize      = eventInfo.EventSize;
    header.boardId        = boardId;
    header.pattern        = eventInfo.Pattern;
    header.triggerTimeTag = eventInfo.TriggerTimeTag;
    header.channelMask    = eventInfo.ChannelMask;
    header.nChannels      = nChannels;
    header.nSamples       = recordLength; // Must all have the same size
                                          // and should match recordLength 
                                          // This check is done below
    header.timeStampSec   = zeit.tv_sec;
    header.timeStampNSec  = zeit.tv_nsec;
    header.fragmentSize   = sizeof(header) 
      + header.nChannels * recordLength * sizeof(uint16_t);

    dataFile->write((char *)&header,sizeof(header));

    retcod = CAEN_DGTZ_DecodeEvent(handle,evtptr,(void **)&event);
    CAENDecoder::checkError(retcod,"DecodeEvent",boardId);

    if ( event->ChSize[0] != recordLength )
    {
      std::cout << CAENDecoder::zeit() 
    		<< " *** nSamples read " << event->ChSize[0]
    		<< " does not match expected "
    		<< recordLength << std::endl;
    }

    for (uint c=0; c<header.nChannels; c++)
    {
      if ( event->ChSize[c] != event->ChSize[0] )
      {
	std::cout << CAENDecoder::zeit() 
		  << " *** BoardId " << boardId 
		  << " Channel " << c 
		  << " nSamples read " << event->ChSize[c]
		  << " does not match first channel "
		  << event->ChSize[0] << std::endl;
      }
      for (uint s=0; s<event->ChSize[c]; s++)
      {
    	dataFile->write((char *)&event->DataChannel[c][s],sizeof(uint16_t));
    	if ( config->runType == LariatReadout::RUNTYPE_PEDESTAL )
    	{
    	  int g = c / MAX_GROUPS;
    	  sum[c]   += (double)event->DataChannel[c][s];
    	  sumsq[c] += (double)event->DataChannel[c][s] * 
    	    (double)event->DataChannel[c][s];
    	  nSamples[c]++;
    	  gsum[g]   += (double)event->DataChannel[c][s];
    	  gsumsq[g] += (double)event->DataChannel[c][s] * 
    	    (double)event->DataChannel[c][s];
    	  nGSamples[g]++;
    	}
      }
    }
  }
  return(eventCounter);
}
