//
//   LariatReadout.cc     (W.Badgett)
//   
//   Class to readout the LArIAT data acquisition cards
//   during a single spill and then return
//
//   This class is truly virtual, a calling or extending class to perform the 
//   actual readout calls to disk and/or artdaq is needed.
//
//   Designed to be a singlet class
//

#include <iostream>
#include <fstream>
#include <ios>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <chrono>
#include "LariatReadout.h"
#include "LariatException.h"
#include "CAENChain.h"
#include "locale.h"
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>

bool LariatReadout::PendingCommand() { return(pendingCommand.load());}
bool LariatReadout::PendingCommand(bool value) 
{ return(pendingCommand.exchange(value));}

void LariatReadout::catchSignal(int signalNumber)
{
  std::cout << "Caught signal " << signalNumber << std::endl;
  if ( signalNumber != SIGPIPE ){  LariatReadout::keepRunning = false;}
}

std::stringstream ifBeamXml;

size_t lariatReadoutCurlCallBack(char* buf, size_t size, size_t nmem, void* up)
{
  for (size_t c = 0; c<size*nmem; c++) { ifBeamXml << buf[c];}
  return(size*nmem);
}

// Destructor:
LariatReadout::~LariatReadout()
{
  runActive = false;
  printNusoft();

  if ( channelScan ) { larasic->testOff();}

  if ( curl != NULL )            { curl_easy_cleanup(curl); curl = NULL;}
  if ( v1740 != NULL )           { delete(v1740); v1740 = NULL; }
  if ( v1751 != NULL )           { delete(v1751); v1751 = NULL; }
  if ( tdc != NULL )             { delete(tdc);   tdc   = NULL; }
  if ( controller != NULL )      { delete(controller); controller = NULL;}
  if ( trigger != NULL )         { delete(trigger);    trigger = NULL;}
  if ( lariatConfig.xml != NULL) { free(lariatConfig.xml); lariatConfig.xml=NULL;}
}

void LariatReadout::fetchIFBeam()
{
  spillOffset    = -1.0;
  spillSnapshot  = -1.0;

  ifBeamXml.clear();
  curl_easy_perform(curl);

  boost::property_tree::ptree tree;
  try
  {
    read_xml(ifBeamXml,tree);
  }
  catch (boost::property_tree::xml_parser::xml_parser_error e )
  {
    std::cout << "Error decoding IFbeam XML" << std::endl;
    std::cout << e.what() << std::endl;
    return;
  }

  // Boost property tree is poorly documented, and some features do not work 
  // in the current implementation, hence the loops and checks.
  //
  // Request is for last one minute;  in principle there could be multiple events, 
  // so look for the most recent one
  //
  try
  {
    boost::property_tree::ptree subtree = tree.get_child("ifbeam-dataset-single-variable");
    boost::property_tree::ptree::const_iterator b0 = subtree.begin();
    boost::property_tree::ptree::const_iterator e0 = subtree.end();
    for ( boost::property_tree::ptree::const_iterator it0= b0; it0!=e0; ++it0)
    {
      if ( it0->first == "variable" )
      {
	boost::property_tree::ptree::const_iterator b1 = it0->second.begin();
	boost::property_tree::ptree::const_iterator e1 = it0->second.end();
	for ( boost::property_tree::ptree::const_iterator it1= b1; it1!=e1; ++it1)
	{
	  if ( it1->first == "<xmlattr>" )
	  {
	    boost::property_tree::ptree::const_iterator b2 = it1->second.begin();
	    boost::property_tree::ptree::const_iterator e2 = it1->second.end();
	    for ( boost::property_tree::ptree::const_iterator it2= b2; it2!=e2; ++it2)
	    {
	      if ( it2->first == "clock" )
	      {
		std::string sSpillSnapshot = it2->second.get_value<std::string>();
		double newSpillSnapshot = atof(sSpillSnapshot.c_str());
		if ( newSpillSnapshot > spillSnapshot )
		{
		  spillSnapshot = newSpillSnapshot;
		  std::string sSpillOffset = it0->second.get_value<std::string>();
		  spillOffset = atof(sSpillOffset.c_str());
		}
	      }
	    }
	  }
	}
      }
    }
  }
  catch (boost::property_tree::ptree_bad_path e )
  {
    std::cout << "Error fetching from IFbeam" << std::endl;
    std::cout << e.what() << std::endl;
    spillOffset = -1.0;
    spillSnapshot = -1.0;
  }
}

sig_atomic_t LariatReadout::keepRunning = true;

void LariatReadout::configure()
{
  if ( v1740 != NULL )
  {
    v1740->configure();
  }

  if ( v1751 != NULL )
  {
    v1751->configure();
  }

  if ( config->runType == RUNTYPE_PHYSICS )       
  { 
    if ( channelScan ) { configurePedestal();}
    else               { configurePhysics(); }
  }
  else if ( config->runType == RUNTYPE_PEDESTAL ) { configurePedestal();}
}

void LariatReadout::lariatReset()
{
  char cmd[256];
  strcpy(cmd,"lariatReset 0 1");
  std::cout << cmd << std::endl;
  system(cmd);
  sleep(4);
}

// Constructor:

LariatReadout::LariatReadout(int runType_, int events_, bool artdaqMode_):
  spillNumber(0),
  artdaqMode(artdaqMode_),
  v1740(NULL),
  v1751(NULL),
  tdc(NULL),
  wut(NULL),
  trigger(NULL),
  controller(NULL),
  larasic(NULL),
  config(NULL),
  spillActive(false),
  runActive(true),
  previousSpillActive(false),
  countV1740(0),
  countV1751(0),
  spillV1740(0),
  spillV1751(0),
  pendingCommand(false),
  channelScan(false)
{
  lariatConfig.xml = NULL;
  curl_global_init(CURL_GLOBAL_ALL);
  curl = curl_easy_init();
  curl_easy_setopt(curl,CURLOPT_URL,
		   "http://ifb-data.fnal.gov:8100/ifbeam/data/data?e=e,21&v=G:E00SNC&tz=&t0=-1m&t1=now&f=xml");
		   
  curl_easy_setopt(curl,CURLOPT_WRITEFUNCTION, &lariatReadoutCurlCallBack);
  curl_easy_setopt(curl,CURLOPT_VERBOSE, 0);

  config = new LariatConfiguration(runType_,events_);

  // Let ctrl-c terminate the run in progress
  LariatReadout::keepRunning = true;
  signal(SIGINT,  catchSignal);
  signal(SIGTERM, catchSignal);
  signal(SIGQUIT, catchSignal);
  signal(SIGPIPE, catchSignal);
  spillNumber = 1;

  if ( config->v1740->enableReadout != 0 )
  {
    v1740 = new CAENChain(config,config->v1740,CAENChain::V1740_CHAIN);
  }
  else { v1740 = NULL;}

  if ( config->v1751->enableReadout != 0 )
  {
    v1751 = new CAENChain(config,config->v1751,CAENChain::V1751_CHAIN);
  }
  else { v1751 = NULL;}

  // config->tdc->enableReadout protects against controllers used only for spillFlag
  if ( config->tdc->enableReadout || config->tdc->spillFlag )
  {
    tdc   = new TDCDriver(config);
    tdc->print();
  }
  else { tdc = NULL;}

  if ( config->larasic->enableReadout )
  {
    larasic = new LARASICDriver(config,config->larasic);
    larasic->print();
    if ( config->larasic->channelScan ) 
    { 
      channelScan = true;
      larasic->testOn();
    }
  }
  else { larasic = NULL;}

  controller = NULL;
  trigger    = NULL;
  wut        = NULL;
  if(config->v2718->enableReadout)
  {
    int32_t handle;
    CAENVME_Init(cvV2718,config->v2718->link,0,&handle);
    controller = new V2718Driver(config,config->v2718,handle);
    controller->configureReadoutRegister();
    controller->print();

    // We talk to V1495 via the V2718, so order is important here
    if(config->v1495->enableReadout)
    {
      trigger = new V1495Driver(config,config->v1495,handle);
      trigger->configureCommon();
      trigger->print();
    }

    // Wave Union TDC (Jinyuan's)
    if ( config->wut->enableReadout )
    {
      wut = new WUTDriver(config,config->wut,handle);
      wut->print();
    }
  }
  else if( (config->v1495->enableReadout) && !(config->v2718->enableReadout))
  {
    throw caen::caen_exception("Cannot enable V1495 without V2718. V1495 needs to be passed in a VME handle from the V2718");
  }

  config->runNumber = fetchRunNumber();
  config->print();
  char xmlFileName[256];
  sprintf(xmlFileName,"/daqdata/lariat_config_%d.xml", config->runNumber);
  std::ofstream xmlFile;
  xmlFile.open(xmlFileName);
  config->concat(xmlFile);
  xmlFile.close();

  uint32_t xmlSize = config->xml.str().size() + 1;
  lariatConfig.fragmentSize = sizeof(lariatConfig.fragmentSize) + 
                              sizeof(lariatConfig.fragmentType) +
                              xmlSize;
  lariatConfig.fragmentType = LariatFragment::FRAGMENT_TYPE_LARIAT_CONFIG;
  lariatConfig.xml = (char *)malloc(xmlSize);
  bzero(lariatConfig.xml,xmlSize);
  strncpy(lariatConfig.xml,config->xml.str().c_str(),xmlSize);

  time_t zeit;
  time(&zeit);
  std::cout << "Starting run " << config->runNumber << " " << ctime(&zeit); 

  if ( config->runType == RUNTYPE_PHYSICS )
  {
    if ( channelScan ) { configurePedestal();}
    else               { configurePhysics();}
  }
  else if ( config->runType == RUNTYPE_PEDESTAL )
  {
    configurePedestal();
  }
}


void LariatReadout::configurePedestal()
{
  if ( v1740 != NULL )
  {
    v1740->configurePedestal();
    v1740->print();
  }

  if ( v1751 != NULL )
  {
    v1751->configurePedestal();
    v1751->print();
  }

  if( trigger != NULL)
  {
    trigger->configurePedestal();
    trigger->print();
  }

  if( controller != NULL)
  {
    controller->configurePedestal();
    controller->print();
  }
  
}

void LariatReadout::configurePhysics()
{
  if ( v1740 != NULL )
  {
    v1740->configurePhysics();
    v1740->print();
  }

  if ( v1751 != NULL )
  {
    v1751->configurePhysics();
    v1751->print();
  }

  if( trigger != NULL){
    trigger->configurePhysics();
    trigger->print();
  }

  if( controller != NULL){
    //    controller->configurePhysics();
    //controller->print();
  }
}

int LariatReadout::fetchRunNumber()
{
  std::ifstream oldRunNumberFile;
  std::ofstream newRunNumberFile;
  char line[100];

  oldRunNumberFile.open("/opt/lariat-online/config/runNumber.dat");
  oldRunNumberFile >> line;
  oldRunNumberFile.close();

  int oldRunNumber = atoi(line);  
  int newRunNumber = oldRunNumber + 1;
  
  newRunNumberFile.open("/opt/lariat-online/config/runNumber.dat");
  newRunNumberFile << newRunNumber;
  newRunNumberFile << std::endl;
  newRunNumberFile.close();
  return(newRunNumber);
}


void LariatReadout::readPedestal(std::ostream * output)
{
  time_t before,after;

  struct LariatFragment::RunInfo runInfo;
  runInfo.fragmentSize         = sizeof(LariatFragment::RunInfo);
  runInfo.fragmentType         = LariatFragment::FRAGMENT_TYPE_RUNINFO;
  runInfo.runNumber            = config->runNumber;
  runInfo.runType              = config->runType;
  runInfo.recordLength         = config->v1740->recordLength;
  runInfo.nBoards              = config->v1740->nBoards;
  runInfo.link                 = config->v1740->link;
  for (int b=0; runInfo.nBoards; b++)
  { runInfo.boardId[b] = config->v1740->boardId[b];}
  runInfo.postPercent          = config->v1740->postPercent;
  runInfo.maxEventsPerTransfer = config->v1740->maxEventsPerTransfer;
  runInfo.eventsPerInterrupt   = config->v1740->eventsPerInterrupt;
  runInfo.irqWaitTime          = config->v1740->irqWaitTime;
  runInfo.usePedestals         = (uint32_t)config->v1740->usePedestals;
  runInfo.dacValue             = config->v1740->dacValue;
  runInfo.samplingRate         = config->v1740->samplingRate;
  runInfo.sampleReduction      = config->v1740->sampleReduction;
  strcpy(runInfo.tdcHostName,config->tdc->hostName.c_str());
  runInfo.tdcPort              = config->tdc->port;
  runInfo.nTDCs                = config->tdc->nBoards;
  runInfo.tdcEnableMask        = config->tdc->enableMask;
  runInfo.tdcNBoards           = config->tdc->nBoards;
  runInfo.tdcTimeOut           = config->tdc->timeOut;
  output->write((char *)&runInfo,sizeof(runInfo));

  output->write((char *)&lariatConfig,sizeof(lariatConfig.fragmentSize)+
		                      sizeof(lariatConfig.fragmentType));
  uint32_t remainder = lariatConfig.fragmentSize - ( sizeof(lariatConfig.fragmentSize)+
						     sizeof(lariatConfig.fragmentType));
  output->write(lariatConfig.xml,remainder);


  countV1740 = 0;
  countV1751 = 0;
  time(&before);

  for ( int e=0; e<config->events && LariatReadout::keepRunning; e++)
  {
    if ( v1740 != NULL )
    {
      v1740->fireSWTrigger();
      v1740->readEvent(output);
    }
    if ( trigger !=NULL)
    {
      trigger->readEvent(output);
    }
    countV1740++;
  }
  LariatReadout::keepRunning = false;
  time(&after);
  time_t duration = after - before;
  std::cout << "Read " << countV1740 << " V1740 triggers in " <<  duration
	    << " seconds" << std::endl;

  if ( v1740 != NULL ) { v1740->writePedestals(output);}

}

void LariatReadout::printHelp()
{
  std::cout << std::endl << "LariatReadout Options:" << std::endl;
  std::cout << "    q   Stop readout and quit" << std:: endl;
  std::cout << "    h   This help screen" << std::endl;
  std::cout << "Every TPC trigger prints 0 (inter-spill) or 1 (during spill)" <<
  std::endl;
  std::cout.flush();
}

//
// This is physics readout, for a single spill
//
// Fills the output ostream and returns the spill number
//

long long LariatReadout::deltaDawn(struct timespec before, 
				   struct timespec after)
{
  long long b = ((long long)before.tv_sec) * 1000000000
    + ((long long)before.tv_nsec);
  long long a = ((long long)after.tv_sec)  * 1000000000
    + ((long long)after.tv_nsec);
  return(a-b);
}

void LariatReadout::writeSpillTrailer(std::ostream * output)
{
  fetchIFBeam();
  struct LariatFragment::SpillTrailer spillTrailer;
  spillTrailer.fragmentSize  = sizeof(spillTrailer);
  spillTrailer.fragmentType  = LariatFragment::FRAGMENT_TYPE_SPILLTRAILER;
  spillTrailer.runNumber     = runNumber()  ;
  spillTrailer.spillNumber   = spillNumber ;
  spillTrailer.timeStamp     = time(NULL);
  spillTrailer.spillSnapshot = spillSnapshot  ; // In milliseconds, time of MidSpill snapshot
  spillTrailer.spillOffset   = spillOffset  ;   // In seconds, normally 2.000 (msec least count)
  spillTrailer.spillV1740    = spillV1740  ;
  spillTrailer.spillV1751    = spillV1751  ;
  spillTrailer.countV1740    = countV1740  ;
  spillTrailer.countV1751    = countV1751  ;
  output->write((char *)&spillTrailer,sizeof(spillTrailer));

  char timeFileName[256];
  sprintf(timeFileName,"/daqdata/lariat_spill_%d_%d.time", spillTrailer.runNumber,
	  spillTrailer.spillNumber);
  std::ofstream timeFile;
  timeFile.open(timeFileName);

  char stupid[80];
  sprintf(stupid,"%12lld", (long long int)spillSnapshot);
  timeFile << stupid << std::endl;
  timeFile.close();
}

int LariatReadout::readSpill(std::ostream * output, std::string fileToCheck)
{

  // JCF, 10/22/14

  // At Kurt's suggestion, have added a feature where if we're in
  // artdaq mode, readSpill will return after timeout_in_secs seconds
  // if no data has been taken

  double timeout_in_secs = 5;

  auto lasttime = std::chrono::high_resolution_clock::now();
  auto lastsize = sizeof(LariatFragment::RunInfo);

  // JCF, 6/13/14

  // Set "artdaqMode" to true if you plan to have this function
  // wrapped by a call to the Lariat fragment generator's "getNext"
  // function


  // JCF, 6/16/14

  // In artdaq, we want "keepRunning" not to mean that readSpill()
  // will never return any more data, but simply that we wish to exit
  // from this function even if the spill isn't officially over as per
  // the TDCs or V1495

  if (artdaqMode) 
  {
    LariatReadout::keepRunning = true;
  }

  struct LariatFragment::RunInfo runInfo;
  runInfo.fragmentSize         = sizeof(LariatFragment::RunInfo);
  runInfo.fragmentType         = LariatFragment::FRAGMENT_TYPE_RUNINFO;
  runInfo.runNumber            = config->runNumber;
  runInfo.runType              = config->runType;
  runInfo.recordLength         = config->v1740->recordLength;
  runInfo.nBoards              = config->v1740->nBoards;
  runInfo.link                 = config->v1740->link;
  for (uint32_t b=0; b<runInfo.nBoards; b++)
  { runInfo.boardId[b]         = config->v1740->boardId[b];}
  runInfo.postPercent          = config->v1740->postPercent;
  runInfo.maxEventsPerTransfer = config->v1740->maxEventsPerTransfer;
  runInfo.eventsPerInterrupt   = config->v1740->eventsPerInterrupt;
  runInfo.irqWaitTime          = config->v1740->irqWaitTime;
  runInfo.usePedestals         = (uint32_t)config->v1740->usePedestals;
  runInfo.dacValue             = config->v1740->dacValue;
  runInfo.samplingRate         = config->v1740->samplingRate;

  output->write((char *)&runInfo,sizeof(runInfo));
  output->write((char *)&lariatConfig,sizeof(lariatConfig.fragmentSize)+
		                      sizeof(lariatConfig.fragmentType));
  uint32_t remainder = lariatConfig.fragmentSize - (sizeof(lariatConfig.fragmentSize)+
						    sizeof(lariatConfig.fragmentType));
  output->write(lariatConfig.xml,remainder);


  int errorCount =0;
  int wutPoll = 0;

  //  controller->pollReadoutRegister(1000000000,1);

  previousSpillActive = false;
  if ( ( controller != NULL) && ( config->v2718->spillFlag ))
  {
    previousSpillActive = controller->spill();
    //previousSpillActive = !( controller->pollReadoutRegister());
  }
  else if ( config->tdc->spillFlag && tdc != NULL )
  {
    previousSpillActive = tdc->spill();
  }
  std::cout << "SpillGate: " << previousSpillActive;
  std::cout.flush();
  if ( config->events <= 0 ) { config->events = 0x7FFFFFFF;}

  // JCF, 6/16/14

  // If we're using this function in the context of artdaq, we'll want
  // to (A) make certain only one spill per call to this function was
  // requested, and (B) relatedly, reset the class member variable
  // "spillNumber" to 1 for each call

  if (artdaqMode) 
  {
    if ( config->events != 1) {
      std::cerr << "Error in LariatReadout: when using artdaq mode, events (i.e., spills) must be set to 1" << std::endl;
      std::exit(1);
    }

    spillNumber = 1;
  }

  for ( int e=0; spillNumber<=config->events && LariatReadout::keepRunning; e++)
  {
    bool haveV1740 = false;
    bool haveV1751 = false;
    char key;
    while ( !haveV1740 && !haveV1751 && LariatReadout::keepRunning)
    {
      // JCF, 12/29/14

      // If the size of the output stream has not increased after
      // timeout_in_secs seconds, that means we haven't gotten any
      // data from the detector; if this is the case (and we're not in
      // a spillActive state) we simply return. This feature allows
      // artdaq's "pause" and "resume" transitions to go through
      // without being blocked

      if ( artdaqMode && !spillActive) {

	auto currenttime = std::chrono::high_resolution_clock::now();

	if (std::chrono::duration_cast<std::chrono::duration<double>>(currenttime-lasttime).count() > timeout_in_secs) {
	  if (static_cast<decltype(lastsize)>(output->tellp()) == lastsize ) {
	    std::cout << "No increase in stream size in last " << timeout_in_secs << " seconds, will exit LariatReadout::readSpill()..." << std::endl;
	    return -1;  // A negative number obviously isn't a spill number!
	  } else {
	    lastsize = output->tellp();
	    lasttime = currenttime;
	  }
	}
      }

      if ( v1740 != NULL ) 
      { 
	if ( channelScan ) { v1740->fireSWTrigger();}
	haveV1740 = v1740->waitForTrigger(); 
      }

      if ( v1751 != NULL ) { haveV1751 = v1751->waitForTrigger();}

      if ( v1740 == NULL && v1751 == NULL ) { usleep(1000); }

      key = 0;
      std::size_t nread = 0;
      if ( ! artdaqMode ) {
	nread = read(STDIN_FILENO, &key, 1);
      } else {

	std::fstream keyfile( fileToCheck );

	if ( keyfile.is_open() ) {
	  
	  // Check and make sure there's only zero or one characters in the file

	  keyfile.seekg( 0, keyfile.end);
	  nread = keyfile.tellg();
	  keyfile.seekg(0, keyfile.beg);

	  if (nread == 1) {
	    keyfile >> key;
	    std::cerr << "LariatReadout: character \"" << key << "\" found in " 
		      << fileToCheck << std::endl;

	    keyfile.close();

	    // JCF, 12/30/14

	    // Blank out the keypress file after reading its character
	    // UNLESS the "q" key was hit, in which case we'll want
	    // the file to continue containing the "q" until the
	    // entire DAQ system is shutdown, which happens at a
	    // higher level than this function

	    if (key != 'q') {
	      std::string cmd = " : > " + fileToCheck;
	      int result = system( cmd.c_str() );

	      if (result != 0) {
		std::cerr << "Warning in LariatReadout: nonzero return value from system() call designed to reset \"" << fileToCheck << "\" to zero bytes" << std::endl;
	      }
	    }

	  } else if (nread > 1) {
	    std::cerr << "Warning in LariatReadout: keystroke file \"" << fileToCheck << "\" is more than one character; will ignore" << std::endl;
	  }
	} else {
	  std::cerr << "Warning in LariatReadout: unable to open keystroke file \"" << fileToCheck << "\"" << std::endl;
	}
      }

      // Pending transition command?  Then return early
      if ( PendingCommand() )
      { 
	std::cout << "PendingCommand is waiting, returning readSpill () early" 
		  << std::endl;
	LariatReadout::keepRunning = false;
      }

      if (nread > 0)
      {
      	switch (key)
      	{
          case 'q': case 'Q': case 3:
      	    LariatReadout::keepRunning = false;
      	    break;
          case 'p': case 'P' :
      	    if ( v1740 != NULL )      v1740->print(); 
      	    if ( v1751 != NULL )      v1751->print(); 
      	    if ( tdc != NULL )        tdc->print();
      	    if ( wut != NULL )        wut->print();
      	    if ( trigger != NULL)     trigger->print();
	    if ( controller != NULL ) controller->print();
      	    std::cout << std::endl; 
      	    break;
          case 'h': case 'H' : case '?' :
      	    printHelp();
      	    break;
      	}
      }

      if ( controller != NULL && config->v2718->spillFlag )
      {
	spillActive = controller->spill();
	//spillActive = !( controller->pollReadoutRegister());
	//	std::cout << spillActive ;  std::cout.flush();
      }
      else if ( tdc != NULL && config->tdc->spillFlag > 0 )
      {
	spillActive = tdc->spill();
      }

      if ( haveV1740 )
      {
	uint32_t eventCounter = 0;
	try
	{
	  eventCounter = v1740->readEvent(output);
	}
	catch ( LariatException e )
	{
	  std::cout << std::endl;
	  e.print();
	  errorCount++;
	  bool keepTrying = true;
	  while ( keepTrying )
	  {
	    try
	    {
	      lariatReset();
	      configure();
	      keepTrying = false;
	    }
	    catch ( LariatException e ) { e.print();}
	  }
	}
	countV1740++;
	spillV1740++;

	if ( larasic != NULL )
	{
	  if ( channelScan )
	  {
	    larasic->writeConfiguration(output,eventCounter);

	    // incrementTestChannel() returns true when all 480 TPC channels 
	    // have been scanned
	    if( larasic->incrementTestChannel() )
	    {
	      keepRunning = false;
	    }
	  }
	  else if ( spillV1740 == 1 )
	  {
	    larasic->writeConfiguration(output,eventCounter);
	  }
	}

	// JCF, 6/16/14

	// The fact that every message in artdaq has a date prepended
	// on screen means that we want to be a little more sparing in
	// our output than if this function was called as part of the
	// LariatRun executable

	if (!artdaqMode || (artdaqMode && countV1740 % 30 == 0) ) 
	{
	  std::cout << spillActive;
	}
	std::cout.flush();
      }

      if ( haveV1751 )
      {
	try
	{
	  v1751->readEvent(output);
	}
	catch ( LariatException e )
	{
	  std::cout << std::endl;
	  e.print();
	  errorCount++;
	  bool keepTrying = true;
	  while ( keepTrying )
	  {
	    try
	    {
	      lariatReset();
	      configure();
	      keepTrying = false;
	    }
	    catch ( LariatException e ) { e.print();}
	  }
	}
	countV1751++;
	spillV1751++;

	if (!artdaqMode || (artdaqMode && countV1751 % 30 == 0) ) 
	{
	  std::cout << spillActive ;
	  std::cout.flush();
	}
      }

      wutPoll++;
      int wutInterval = 0;
      if ( config->wut->readoutPeriod > 0 )
      { wutInterval = wutPoll % config->wut->readoutPeriod;}

      if ( (wutInterval == 0 ) && ( wut!=NULL) )
      {
	wut->readEvent(output,wutPoll);
	std::cout.flush();
	printNusoft();
      }


      // Now we need to check if the end-of-spill has occurred 
      // We have two possible sources of the spill flag:  the Wire 
      // Chamber TDC controller or the VME V2718 crate controller, 
      // so this is a skosh complicated.   Preferentially we will 
      // use the V2718, since this is the normal mode of Lariat 
      // operation.

      // The spillIsDone flag should only be true if we 
      // observe a TRANSITION from "spill on" to "spill off"
      bool spillIsDone = false;

      // Is the V2718 controller present?   If yes, use it!
      if ( ( controller != NULL) && ( config->v2718->spillFlag ))
      {
	//spillActive = !( controller->pollReadoutRegister());
        spillActive = controller->spill();
      }
      else if ( ( tdc != NULL ) && ( config->tdc->spillFlag ) )
      {
	spillActive = tdc->spill();
      }

      // static uint64_t loop_cntr = 0;

      // if (loop_cntr++ % 100 == 0) {
      //      std::cout << "Spill active status is " << spillActive << ", output stream is " << output->tellp() << std::endl;
      // }

      if ( spillActive != previousSpillActive )
      {
	if ( previousSpillActive ) 
	{ 
	  spillIsDone = true;
	}
	previousSpillActive = spillActive;
      }

      // If we have really found the end of spill, then we need to 
      // read out trigger, WUT, W/C TDCs, close the data file and 
      // write out the web status files:

      if(spillIsDone) 
      {
	if ( trigger != NULL )
	{
	  trigger->readEvent(output);
	  // trigger->getFragment().print();

	  std::ofstream of0("/nusoft/app/web/htdoc/lariat/channels.html");
	  std::ofstream of1("/nusoft/app/web/htdoc/lariat/triggers.html");
	  std::ofstream of2("/nusoft/app/web/htdoc/lariat/vetos.html");
	  trigger->getFragment().printHTML(0, of0);
	  trigger->getFragment().printHTML(1, of1);
	  trigger->getFragment().printHTML(2, of2);

	  printNusoft();
	}

	// we rely on counters being reset prior to each beam spill by
	// a hardware input
	if ( controller != NULL )
	{
	  controller->resetReadoutRegister();
	}

	// The WUT is asynchronous and double buffered, so we need to 
	// read out any straggler hits left from after the last time 
	// we read it:
	if ( wut != NULL ) 
	{ 
	  wut->readEvent(output,wutPoll);
	  wutPoll++;
	}

	// Could be left-over triggers in the V1740s and V1751
	if ( v1740 != NULL )
	{
	  while ( v1740->readEvent(output) > 0 ) 
	  {
	    countV1740++;
	    spillV1740++;
	  }
	}

	if ( v1751 != NULL )
	{
	  while ( v1751->readEvent(output) > 0 ) 
	  {
	    countV1751++;
	    spillV1751++;
	  }
	}

	// The W/C TDC system is designed to cache all hits during 
	// a spill in memory, and only read out when the spill 
	// is over, so we read the whole dang thing now:
	if ( ( tdc != NULL ) && 
	     config->tdc->enableReadout &&
	     ( config->tdc->nBoards > 0 ) )
	{
	  tdc->readEvent(output);
	}

	if (! artdaqMode ) 
	{
	  std::cout << std::endl << "Spill " << spillNumber
		    << " has ended" << std::endl;
	  std::cout << std::endl << "Read " << spillV1740 << "/" <<
	    countV1740 <<  " V1740 triggers and "
		    << spillV1751 << "/" << 
	    countV1751 << " V1751 triggers" << std::endl;
	} 
	else
	{
	  std::cout << std::endl << "Spill has ended" << std::endl;
	}
	writeSpillTrailer(output);

	spillV1740 = 0;
	spillV1751 = 0;

	spillNumber++;
	wutPoll = 0;
      
	// ArtDaq expects us to return every spill, and has a 
	// separate command mechanism to stop running
	if ( !artdaqMode && ( spillNumber > config->events ) )
	{ 
	  LariatReadout::keepRunning = false;
	}
	std::cout.flush();

	// Ignore spill if we're doing an ASIC channel scan
	if ( channelScan )
	{
	  std::cout << "ASIC Channel Scan in progress, ignoring end of spill"
		    << std::endl;
	}
	else
	{
	  return(spillNumber);
	}
      }
    }
  }

  if ( !artdaqMode && ( spillNumber > config->events ) )
  { LariatReadout::keepRunning = false;}

  std::cout << std::endl << "Read " << countV1740 << " V1740 triggers and "
	    << countV1751 << " V1751 triggers" << std::endl;
  return(spillNumber);
}


void LariatReadout::printNusoft()
{
  std::ofstream of;

  of.open("/nusoft/app/web/htdoc/lariat/run.html.new");
  printHTML(of);
  of.close();
  system("mv /nusoft/app/web/htdoc/lariat/run.html.new /nusoft/app/web/htdoc/lariat/run.html");
}

void LariatReadout::printHTML(std::ostream & os)
{
  os << "<html>\n\
  <head>\n\
  <title>LArIAT Run Status</title>\n\
  <link type=\"text/css\" rel=\"stylesheet\" href=\"runStyles.css\"/>\n\
  <meta http-equiv=\"refresh\" content=\"2\">\
  </head>\
  <body>\n\
  <table align=\"center\" border=\"0\">\n\
  <tr>\n\
  <td>\n\
    <table align=\"center\" border=\"0\" cellpadding=\"3\">\n\
    <tr>\n\
    <td>&nbsp;</td>\n\
   <td align=center style=\"font-weight:bold;font-family:sans-serif;color:#4499aa;font-size:200%\">LArIAT Run Status\
 <td>&nbsp;</td>\n\
</tr>\n\
</table>" << std::endl;

  os << "<BR><BR><table width=560 align=\"center\" border=\"0\" cellpadding=\"3\">" << std::endl;
  os << "<tr><td align=right><B>Run</B></td><td>" << runNumber() << "</td><td align=right><B>Spill</B></td><TD>" <<
    spillNumber << "</td></tr>" << std::endl;
  os << "<tr><td align=right><B>V1740 Run Triggers</B></td><td>" << countV1740 << 
    "<td align=right><B>V1740 Spill Triggers</B></td><td>" << spillV1740 << "</td></tr>" << std::endl;
  os << "<tr><td align=right><B>V1751 Run Triggers</B></td><td>" << countV1751 << 
    "<td align=right><B>V1751 Spill Triggers</B></td><td>" << spillV1751 << "</td></tr>" << std::endl;

  os << "<tr><td align=right><B>SpillActive</B></TD><TD>" << spillActive << "</TD><TD align=right><B>RunActive</B></td><td>"<<runActive<<"</td></tr>"<< std::endl;

  os << "<tr><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td></tr>" << std::endl;
  if ( trigger != NULL )
  {
    V1495Fragment t = trigger->getFragment();
    if ( t.channelData.size() > 0 )
    {
      for(size_t i=0; i<t.channelData.size(); ++i)
      {
	os << "<tr><td align=right><B>" << t.channelData[i].name << "</B></td><td>" 
	   << t.channelData[i].counts << "</td><td align=right><B>" << t.channelData[i+1].name
	   << "</b></td><td>" << t.channelData[i+1].counts << "</td></tr>" << std::endl;
	++i;
      }
    }
    else
    {
      V1495InputChannels cc = config->v1495->inputChannels;
      for(size_t i=0; i<cc.size(); ++i)
      {
	os << "<tr><td align=right><B>" << cc[i].name << "</B></td><td>" 
	   << cc[i].counts << "</td><td align=right><B>" << cc[i+1].name
	   << "</B></td><td>" << cc[i+1].counts << "</td></tr>" << std::endl;
	++i;
      }
    }
  }

  os << "</table><P><P>" << std::endl;
  os << "</body>" << std::endl;
  os << "</html>" << std::endl;
}

