//
//  TDCDriver.cc  (W.Badgett)
//
//  Class to configure and readout one of a chain
//  of CAEN TDC wave form digitizers
//

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <ctime>
#include <errno.h>
#include <netdb.h>
#include <time.h>
#include <exception>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <byteswap.h>
#include "TDCDriver.h"
#include "TDCFragment.h"
#include "LariatFragment.h"
#include "LariatConfiguration.h"

TDCDriver::~TDCDriver()
{
  int retcod;
  std::cout << "Closing socket" << std::endl;
  retcod = shutdown(localSocket,2);
  if ( retcod < 0 )
  {
    std::cout << "Error on socket shutdown "<< strerror(errno) << std::endl;
    return;
  }

  retcod = close(localSocket);
  if ( retcod < 0 )
  {
    std::cout << "Error on socket close "<< strerror(errno) << std::endl;
  }

  free(buffer);
}

char * TDCDriver::zeit()
{
  time_t t;
  struct tm * hora;
  static char heure[20];

  std::time(&t);
  hora = std::localtime(&t);
  std::strftime(heure,sizeof(heure),"%Y.%m.%d %H:%M:%S",hora);
  return((char *)heure);
}

TDCDriver::TDCDriver(LariatConfiguration *config_)
{
  int retcod;

  config = config_;
  ok = true;
  buffer = (uint16_t *)malloc(CONTROLLER_MEMORY_SIZE);

  localSocket = socket(AF_INET, SOCK_STREAM, 0);
  if ( localSocket < 0 )
  {
    std::cout << "Error opening socket " << strerror(errno) << std::endl;
    ok = false;
    return;
  }

  remoteHost = gethostbyname(config->tdc->hostName.c_str());
  if ( remoteHost == NULL )
  {
    std::cout << "Error finding host " << config->tdc->hostName << " " 
	      << strerror(errno) << std::endl;
    ok = false;
    return;
  }

  remoteSocket.sin_family = remoteHost->h_addrtype;
  remoteSocket.sin_port   = htons(config->tdc->port);
  remoteSocket.sin_addr   = *((struct in_addr *) remoteHost->h_addr ) ;

  char *addressStringPtr = inet_ntoa(remoteSocket.sin_addr);
  strcpy(addressString, addressStringPtr);

  retcod = connect(localSocket,
		   (const sockaddr*)&remoteSocket,sizeof(remoteSocket));
  if ( retcod < 0 )
  {
    std::cout << "Error connecting to " << config->tdc->port << 
      "@" <<  config->tdc->hostName << " " 
      << strerror(errno) << std::endl;
    ok = false;
    return;
  }

  struct timeval timeOut;
  timeOut.tv_sec  = config->tdc->timeOut;
  timeOut.tv_usec = 0;
  setsockopt(localSocket, SOL_SOCKET, SO_RCVTIMEO, (char *)&timeOut, 
	     sizeof(struct timeval));

  if ( config->tdc->enableMask )
  {
    // Reset TDC system.
    uint16_t data = ( 1 << SEND_2_FM_TX_ENABLE_S ) | 
      //      ( 1 << PLL_REFERENCE_SOURCE_S ) | 
      ( 1 << RESET_SPILL_COUNTER_S )  |
      ( 1 << FIFO_BUFFER_RESET_S )    |
      ( 1 << READOUT_SEQUENCER_RESET_S );
    write(CSR_ADDR, data);

    // Specify TCLK start and end spill events
    write(START_SPILL_ADDR,config->tdc->startSpill);
    write(END_SPILL_ADDR,  config->tdc->endSpill);

    if ( config->tdc->pulserEnable )
    {
      data = read(CSR_ADDR);
      data |= ( 1 << TEST_PULSER_ENABLE_S);
      write(CSR_ADDR,data);
      write(PULSER_FREQUENCY_LOW_ADDR,(config->tdc->pulserFrequency & 0xFFFF));
      write(PULSER_FREQUENCY_HIGH_ADDR,((config->tdc->pulserFrequency>>16) & 0xFFFF));
      write(PULSER_SPILL_WIDTH_ADDR,config->tdc->spillWidth);
      write(PULSER_INTERSPILL_WIDTH_ADDR,config->tdc->interSpillWidth);
    }
    else
    {
      write(PULSER_FREQUENCY_LOW_ADDR,(uint16_t)0);
      write(PULSER_FREQUENCY_HIGH_ADDR,(uint16_t)0);
      write(PULSER_SPILL_WIDTH_ADDR,(uint16_t)0);
      write(PULSER_INTERSPILL_WIDTH_ADDR,(uint16_t)0);
    }

    // Now configure all TDCs involved in readout
    uint32_t csr = TDC_TDC_CSR00_NOMINAL | 
      ( config->tdc->ignoreSpill  << TDC_TDC_IGNORE_SPILL_S ) | 
      ( config->tdc->outputSelect << TDC_TDC_OUTPUT_SELECT_S ) ;
    write(TDC_TDC_CSR_ADDR,csr,0);
    write(TDC_TDC_GATE_WIDTH_ADDR,config->tdc->gateWidth,0);
    write(TDC_TDC_INITIAL_TIMESTAMP_ADDR,TDC_TDC_INITIAL_TIMESTAMP_V,0);
    write(TDC_TDC_PIPELINE_DELAY_ADDR,config->tdc->pipeLineDelay,0);
    for ( int i=0; i<TDCConfiguration::N_CHANNEL_MASKS; i++)
    {
      write(TDC_TDC_CHANNEL_ENABLE_BASE_ADDR+i,
	    TDC_TDC_ALL_CHANNELS,0);
    }
    for ( int i=0; i<TDCConfiguration::N_THRESHOLD_REGISTERS; i++)
    {
      write(TDC_TDC_THRESHOLD_BASE_ADDR+i,config->tdc->threshold,0);
    }
  }
}

int TDCDriver::countBits(uint16_t mask)
{
  int reply = 0;
  for ( uint i=0; i<sizeof(uint16_t)*8; i++)
  { if ( ( 1 << i ) & mask ) { reply++;} }
  return(reply);
}

int TDCDriver::maxBit(uint16_t mask)
{
  int reply = 0;
  for ( uint i=sizeof(uint16_t)*8-1; i>=0 && reply==0 ; i--)
  { if ( ( 1 << i ) & mask ) { reply = i + 1;} }
  return(reply);
}

int TDCDriver::minBit(uint16_t mask)
{
  int reply = 0;
  for ( uint i=0; i<sizeof(uint16_t)*8 && reply==0 ; i++)
  { if ( ( 1 << i ) & mask ) { reply = i+1;} }
  return(reply);
}


bool TDCDriver::spill()
{
  uint16_t data = read(SPILL_STATE_ADDR);
  bool spillFlag = (data >> SPILL_GATE_S ) & 0x1;
  return(spillFlag);
}

void TDCDriver::write(uint32_t address, uint32_t data, int32_t tdc)
{
  char cmd[256];

  if ( tdc < 0 ) { sprintf(cmd, "wr %x %x\n", address, data);}
  else           { sprintf(cmd, "hc %d wr %x %x\n", tdc, address, data);}

  int retcod = send(localSocket, cmd, strlen(cmd), 0);
  if ( retcod <= 0 )
  {
    std::cout << "Error on send socket " << strerror(errno) << std::endl;
    return;
  }
}

uint32_t TDCDriver::read(uint32_t address, int32_t tdc)
{
  uint32_t data;
  char cmd[256];
  char reply[65536];
  //int responseLength;
  int length=0,size=0,p=0;
  
  // avoid set but not used compiler warning
  length = size;
  size = length;

  if ( tdc < 0 ) { sprintf(cmd, "rd %x\n", address);}
  else           { sprintf(cmd, "hc %d rd %x\n", tdc, address);}

  int retcod = send(localSocket, cmd, strlen(cmd), 0);
  if ( retcod <= 0 )
  {
    std::cout << "Error on send socket " << strerror(errno) << std::endl;
    return(-1);
  }

  data = 0; p = 1;
  size = sizeof(reply);
  bzero(reply,sizeof(reply));
  reply[0] = '0';
  while ( reply[p-1] != 10 )
  {
    length = recv(localSocket,&reply[p],sizeof(char),0);
    p++;
  }

  if ( p > 0 ) { sscanf(reply,"%x", &data); }
  else         { data = -1; }
  return(data);
}

void TDCDriver::print(std::ostream & os)
{
  os << "TDC HostName " << config->tdc->hostName 
	    << " Port " << config->tdc->port << std::endl;
  os << "  nBoards    " << config->tdc->nBoards << std::endl;
  readConfiguration();
  os << "  CSR            0x" << std::hex << csr << std::endl;
  os << "  TClk Start     0x" << tclkStart << std::endl;
  os << "  TClk End       0x" << tclkEnd << std::dec << std::endl;
  os << "  TriggerCounter      " << triggerCounter << std::endl;
  os << "  SpillWordCount      " << spillWordCount << std::endl;
  os << "  TDCSpillStatus 0x" << std::hex << tdcSpillStatus << std::endl;
  os << "  TDCEnabled     0x" << tdcEnabled << std::dec << std::endl;
  os << "  SpillCounter        " << spillCounter << std::endl;
  os << "  SpillState     0x" << std::hex << spillState << std::dec << std::endl;
  for (int i=0; i<TDCFragment::MAX_TDCS; i++)
  {
    os << "  TDC " << i << " WordCount   " << tdcWordCount[i] <<
      "    LinkStatus 0x" << std::hex << tdcLinkStatus[i] <<"   InterruptStatus 0x" << 
      tdcInterruptStatus[i] << " Spill " << std::dec << tdcSpill[i] << std::endl;
  }
  os << "  PulserFreqency     " << pulserFrequency << " " << 
    ((double)pulserFrequency * 0.0247) << " Hz" << std::endl;
  os << "  SpillGate          " << spillGate << std::endl;
  os << "  EOTComplete        " << eotComplete << std::endl;
  os << "  TDCReadoutDone     " << tdcReadoutDone << std::endl;
  os << "  TDCReadoutBusy     " << tdcReadoutBusy << std::endl;
  os << "  SingleStep         " << singleStep << std::endl;
  os << "  TestPulserEnabled  " << testPulserEnabled << std::endl;
  os << "  PLLReferenceSource " << pllReferenceSource << std::endl;
  os << "  DMABusy            " << dmaBusy << std::endl;
  os << "  SpillWidth         " << spillWidth << std::endl;
  os << "  InterSpillWidth    " << interSpillWidth << std::endl;
}

bool TDCDriver::checkTdcNumber(uint8_t n, uint8_t expected)
{
  bool bad = false;
  if ( n != expected )
  {
    std::cout << "*** " << zeit() << 
      " Found wrong TDC number: " << (int)n << " expected " << 
      (int)expected << std::endl;
    bad = true;
    //    throw std::range_error("TDC number wrong");
  }
  return(bad);
}

void TDCDriver::readEvent(std::ostream *dataFile)
{
  char cmd[256];
  TDCFragment::TdcControllerHeader header;
  TDCFragment::TdcSpillHeader tdcSpill;
  TDCFragment::TdcEventHeader tdcEvent;
  TDCFragment::TdcHit hits[TDCFragment::MAX_HITS];

  bool tdcBusy = true;
  bool tdcDone = false;
  bool eotComplete = false;
  int ctr = 0;
  struct timespec wait, done;

  clock_gettime(CLOCK_REALTIME, &wait);

  while ( !tdcDone && !eotComplete && tdcBusy && ( ctr<1000))
  {
    uint16_t data = read(SPILL_STATE_ADDR);
    tdcBusy = ( data >> TDC_READOUT_BUSY_S ) & 0x1;
    tdcDone = ( data >> TDC_READOUT_DONE_S ) & 0x1;
    eotComplete = ( data >> EOT_COMPLETE_S ) & 0x1;
    ctr++;
  }

  if (tdcBusy)
  {
    std::cout << "Timeout waiting on TDC Done" << std::endl;
    return;
  }

  clock_gettime(CLOCK_REALTIME, &done);
  dataSize = getSpillWordCount();
  if ( config->tdc->debugLevel >= 1 ) 
  { 
    unsigned long long w = wait.tv_sec * 1000000000 + wait.tv_nsec;
    unsigned long long d = done.tv_sec * 1000000000 + done.tv_nsec;
    unsigned long long delta = d - w;

    std::cout << "Got TDC event of size " << dataSize <<
      " in " << ( delta / 1000000000 ) << " , " << ( delta % 1000000000 ) 
	      << std::endl;
  }
  uint32_t dataBytes = dataSize * sizeof(uint16_t);
  // uint32_t dataBytes = dataSize;
  if ( config->tdc->debugLevel >= 5 ) { print();}

  sprintf(cmd, "rdb\n");
  int retcod = send(localSocket, cmd, strlen(cmd), 0);
  if ( retcod <= 0 )
  {
    std::cout << "Error on send socket " << strerror(errno) << std::endl;
    return;
  }

  char * cbuffer = (char *)buffer;
  if ( dataBytes > 0 )
  {
    uint32_t dataRead = 0;
    bool areOK = true;
    while ( ( dataRead < dataBytes  ) && areOK)
    {
      retcod = recv(localSocket,(char *)&cbuffer[dataRead],(dataBytes-dataRead),0);
      if ( retcod <= 0 )
      {
	std::cout << "Error on recv socket after " << dataRead << " bytes " <<
	  strerror(errno) << " [" << errno << "]" 
		  << std::endl;
	ok = false;
	areOK = false;
      }
      else
      {
	dataRead += retcod;
      }
    }
  }
  else
  {
    std::cout << "No data found from TDCs" << std::endl;
    return;
  }

  // Unpack the binary stream
  header.fragmentSize = sizeof(header);
  header.fragmentType = LariatFragment::FRAGMENT_TYPE_TDC_CONTROLLER_HEADER;

  ptr = 0;
  header.spillWordCount = dataSize;
  header.spillWordCountBuffer = ( bswap_16(buffer[ptr]) << 16 ) | bswap_16(buffer[ptr+1]) ;
  if ( header.spillWordCount != header.spillWordCountBuffer )
  {
    std::cout << "*** " << zeit() << " Spill word counts do not agree, register " 
	      << header.spillWordCount << " buffer " 
	      << header.spillWordCountBuffer << std::endl;
    return;
  }
  incr(); incr();
  header.spillCounter = bswap_16(buffer[ptr]);
  incr();
  header.month   = ( buffer[ptr] >> 8 ) & 0xFF;
  header.year    =   buffer[ptr]      & 0xFF;;
  incr();
  header.hours   = ( buffer[ptr] >> 8 ) & 0xFF;;
  header.day     =   buffer[ptr]      & 0xFF;
  incr();
  header.seconds = ( buffer[ptr] >> 8 ) & 0xFF;
  header.minutes =   buffer[ptr]        & 0xFF;
  incr();
  header.triggerCounts = ( bswap_16(buffer[ptr]) << 16 ) | bswap_16(buffer[ptr+1]);
  incr(); incr();
  header.tdcStatus   =   buffer[ptr];
  if ( header.tdcStatus )
  {
    std::cout << zeit() << " *** Unexpected Comtroller TDC Status word 0x" << 
      std::hex << header.tdcStatus << std::dec << std::endl;
  }
  incr();
  header.linkStatus  =   buffer[ptr];
  if ( header.linkStatus )
  {
    std::cout << zeit () << " *** Unexpected Comtroller Link Status word 0x" << 
      std::hex <<  header.tdcStatus << std::dec << std::endl;
  }
  incr();
  // Not sure why two reads are necessary
  header.tdcActiveMask = read(LINK_ACTIVITY_ADDR);
  header.tdcActiveMask = read(LINK_ACTIVITY_ADDR);
  if ( header.tdcActiveMask != config->tdc->enableMask )
  {
    std::cout << zeit() << 
      " *** Unexpected TDC mask 0x" << std::hex << header.tdcActiveMask <<
      ", expected 0x" << config->tdc->enableMask << std::dec << std::endl;
  }

  header.controllerNumber = config->tdc->controllerNumber;
  header.nTDCs = countBits(header.tdcActiveMask);

  // BJR 10/29/14 comment out unused variables
  //uint16_t minTdc = minBit(header.tdcActiveMask);
  //uint16_t maxTdc = maxBit(header.tdcActiveMask);
 
  dataFile->write((char *)&header, sizeof(header));
  LariatFragment::checkOStream(*dataFile);
  //  if ( config->tdc->debugLevel >= 6 ) { TDCFragment::print(&header);}

  uint counters[header.nTDCs];
  for ( uint i=0; i<(header.nTDCs) && ((uint)ptr<dataSize); i++)
  {
    tdcSpill.fragmentSize = sizeof(tdcSpill);
    tdcSpill.fragmentType = LariatFragment::FRAGMENT_TYPE_TDC_TDC_SPILL_HEADER;
    tdcSpill.spillWordCount = ( bswap_16(buffer[ptr]) << 16 ) | bswap_16(buffer[ptr+1]);
    incr(); incr();
    tdcSpill.controllerNumber = config->tdc->controllerNumber;
    tdcSpill.tdcNumber = ( buffer[ptr] >> 8 ) & 0xFF;

    incr();
    tdcSpill.spillTriggerCounter = ( bswap_16(buffer[ptr]) << 16 ) 
      | bswap_16(buffer[ptr+1]);
    incr(); incr();
    tdcSpill.spillStatus = bswap_16(buffer[ptr]);
    if ( tdcSpill.spillStatus != GOOD_TDC_STATUS )
    {
      std::cout << zeit() << " *** Bad TDC " << tdcSpill.tdcNumber << 
	" spill status 0x" << std::hex << tdcSpill.spillStatus << 
	", expected 0x" << GOOD_TDC_STATUS << std::dec << std::endl;
    }
    incr();
    dataFile->write((char *)&tdcSpill,sizeof(tdcSpill));
    LariatFragment::checkOStream(*dataFile);
    //    if ( config->tdc->debugLevel >= 6 ) { TDCFragment::print(&tdcSpill);}
    checkTdcNumber(tdcSpill.tdcNumber,(char)(i+1));
    counters[i] = tdcSpill.spillTriggerCounter;
  }

  bool counterMismatch = false;
  for ( uint i=1; i<header.nTDCs; i++)
  { if ( counters[i] != header.triggerCounts ) { counterMismatch = true;}   }
  if ( counterMismatch )
  {
    std::cout << zeit() << "*** TDC Total Spill Trigger Counter Mismatch" << std::endl;
    std::cout << "*** Counters: " ;
    for ( uint i=0; i<header.nTDCs; i++)
    {  std::cout << counters[i] << " " ;}
    std::cout << "Controller: " << header.triggerCounts << std::endl;
    return;
    //throw std::length_error("TDC Counter Mismatch");
  }

  for ( uint i=0; i<header.triggerCounts; i++)
  {
    bool counterMismatch = false;
    for ( uint j=0; j<header.nTDCs; j++)
    {
      tdcEvent.controllerNumber = config->tdc->controllerNumber;
      tdcEvent.fragmentType   = LariatFragment::FRAGMENT_TYPE_TDC_TDC_EVENT_HEADER;
      tdcEvent.eventWordCount = ( buffer[ptr] >> 8 ) & 0xFF;
      eventSize = tdcEvent.eventWordCount;
      eptr = 0;
      incre();
      tdcEvent.tdcNumber      = ( buffer[ptr] >> 8 ) & 0xFF;
      checkTdcNumber(tdcEvent.tdcNumber,(char)(j+1));
      incre();
      tdcEvent.eventStatus    = ( buffer[ptr] >> 8 ) & 0xFF;
      incre();
      tdcEvent.triggerCounter = ( bswap_16(buffer[ptr]) << 16) | bswap_16(buffer[ptr+1]);
      counters[j]             = (int)tdcEvent.triggerCounter;
      incre(); incre();
      tdcEvent.triggerType    = ( buffer[ptr] >> 8 ) & 0xF;
      incre();
      tdcEvent.controllerTimeStamp = bswap_16(buffer[ptr]) & 0xFFF;
      incre();
      tdcEvent.tdcTimeStamp   = ( bswap_16(buffer[ptr]) << 16) | bswap_16(buffer[ptr+1]);
      incre(); incre();

      // Check consistency between two time stamps
      uint32_t zeit0 = ((uint32_t)tdcEvent.controllerTimeStamp) >> 3;
      uint32_t zeit1 = tdcEvent.tdcTimeStamp & 0x1FF;
      if ( zeit0 != zeit1 ) 
      {
	std::cout << zeit () << 
	  " *** TDC " << j << " Controller time bits 0x" << std::hex << zeit0 << 
	  " mismatch TDC time bits 0x" << zeit1 << " [0x" << tdcEvent.controllerTimeStamp<<
	  ",0x" << tdcEvent.tdcTimeStamp <<"]" << std::endl;
      }

      if ( TDCFragment::TDC_EVENT_HEADER_SIZE <= tdcEvent.eventWordCount )
      {
	tdcEvent.nHits = tdcEvent.eventWordCount - TDCFragment::TDC_EVENT_HEADER_SIZE;
      }
      else 
      { 
	tdcEvent.nHits = 0;
	std::cout << zeit() << " *** Illegal TDC Event Size Word Count " << 
	  (int)tdcEvent.eventWordCount << std::endl;
      }

      tdcEvent.fragmentSize   = sizeof(tdcEvent) + 
	tdcEvent.nHits * sizeof(TDCFragment::TdcHit);
      dataFile->write((char *)&tdcEvent,sizeof(tdcEvent));
      LariatFragment::checkOStream(*dataFile);

      //      if ( config->tdc->debugLevel >= 6 ) { TDCFragment::print(&tdcEvent);}
      for ( uint j = 0; (j<tdcEvent.nHits) && ((uint)ptr < dataSize) ; j++)
      {
	hits[j].channel = ( buffer[ptr] >> 2 ) & 0x3F;
	hits[j].timeBin = ( ( buffer[ptr] >> 8 ) & 0xFF ) |
	  ( ( buffer[ptr] << 8 ) &  0x300 );
	incre();
	//	if ( config->tdc->debugLevel >= 6 ) { TDCFragment::print(&hits[j], j);}
      }
      int size = tdcEvent.nHits * sizeof(TDCFragment::TdcHit);
      dataFile->write((char *)hits, size);
      LariatFragment::checkOStream(*dataFile);
    }

    for ( uint j=1; j<header.nTDCs; j++)
    { if ( counters[j] != (i+1) ) { counterMismatch = true;}   }
    if ( counterMismatch )
    {
      std::cout << zeit() << " *** TDC Trigger Number Mismatch in Event" << 
	std::endl;
      std::cout << "*** Counters: " ;
      for ( uint j=0; j<header.nTDCs; j++)
      {  std::cout << counters[j] << " " ;}
      std::cout << ", expected " << (i+1) << std::endl;
      //throw std::length_error("TDC Event Counter Mismatch");
    }
  }
  if ( config->tdc->debugLevel >= 3 ) 
  {
    std::cout << "Actual size " << ptr << " header size " << 
      dataSize << std::endl;
    std::cout << std::endl;
  }
}

uint32_t TDCDriver::getSpillWordCount()
{
  spillWordCount = ( read(SPILL_WORD_COUNT_HIGH_ADDR) << 16 ) |
    read(SPILL_WORD_COUNT_LOW_ADDR);
  return(spillWordCount);
}

void TDCDriver::readConfiguration()
{
  csr = read(CSR_ADDR);
  tclkStart = read(START_SPILL_ADDR);
  tclkEnd = read(END_SPILL_ADDR);
  triggerCounter = ( read(SPILL_TRIGGER_COUNTER_HIGH_ADDR) << 16 ) |
                     read(SPILL_TRIGGER_COUNTER_LOW_ADDR);
  spillWordCount = ( read(SPILL_WORD_COUNT_HIGH_ADDR) << 16 ) |
                     read(SPILL_WORD_COUNT_LOW_ADDR);
  pulserFrequency = ( read(PULSER_FREQUENCY_HIGH_ADDR) << 16 ) |
                     read(PULSER_FREQUENCY_LOW_ADDR);
  tdcSpillStatus = read(TDC_SPILL_STATUS_ADDR);
  spillCounter   = read(SPILL_COUNTER_ADDR);
  spillState     = read(SPILL_STATE_ADDR);
  tdcEnabled     = read( LINK_ACTIVITY_ADDR);
  for ( int i=0; i<TDCFragment::MAX_TDCS; i++)
  {
    tdcWordCount[i]       = read(TDC_WORD_COUNT_BASE_ADDR+i);
    tdcLinkStatus[i]      = read(TDC_PORT_LINK_STATUS_BASE_ADDR+i);
    tdcInterruptStatus[i] = read(TDC_PORT_INTERRUPT_STATUS_BASE_ADDR+i);
    tdcSpill[i]           = read(TDC_TDC_SPILL_STATUS_ADDR,(i+1));
  }
  spillGate            = (spillState >>  SPILL_GATE_S ) & 1;
  eotComplete          = (spillState >> EOT_COMPLETE_S) & 1;
  tdcReadoutDone       = (spillState >> TDC_READOUT_DONE_S) & 1;
  tdcReadoutBusy       = (spillState >>TDC_READOUT_BUSY_S  ) & 1;
  singleStep           = (csr >>SINGLE_STEP_S ) & 1;
  testPulserEnabled    = (csr >>TEST_PULSER_ENABLE_S ) & 1;
  pllReferenceSource   = (csr >>PLL_REFERENCE_SOURCE_S ) & 1;
  spillWidth           = read(PULSER_SPILL_WIDTH_ADDR);
  interSpillWidth      = read(PULSER_INTERSPILL_WIDTH_ADDR);
}


