//
//  V1740Driver.cc  (W.Badgett)
//
//  Class to configure and readout one of a chain
//  of CAEN V1740 wave form digitizers
//

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <ctime>
#include <ptree.hpp>
#include "V1740Driver.h"
#include "CAENFragment.h"
#include "LariatConfiguration.h"
#include "LariatFragment.h"
#include "LariatReadout.h"
#include "CAENDecoder.h"

void V1740Driver::configureCommon()
{
  CAEN_DGTZ_ErrorCode retcod = CAEN_DGTZ_Success;
  CAENDriver::configureCommon();

  for (group=0; group<CAENDriver::MAX_GROUPS; group++)
  {
    channelMask[group]          = 0xFF; 
    threshold[group]            = 0;
    groupSelfTriggerMode[group] = CAEN_DGTZ_TRGMODE_DISABLED;
    offset[group]               = config->v1740->dacValue; 

    CAEN_DGTZ_SetChannelGroupMask(handle,group,channelMask[group]);
    CAENDecoder::checkError(retcod,"SetChannelGroupMask",boardId);
    CAEN_DGTZ_SetGroupTriggerThreshold(handle,group,threshold[group]);
    CAENDecoder::checkError(retcod,"SetGroupTriggerThreshold",boardId);
    CAEN_DGTZ_SetGroupDCOffset(handle,group,offset[group]);
    CAENDecoder::checkError(retcod,"SetGroupDCOffset",boardId);
    CAEN_DGTZ_SetGroupSelfTrigger(handle,groupSelfTriggerMode[group],group);
    CAENDecoder::checkError(retcod,"SetGroupSelfTrigger",boardId);
  }

  // Should be CAEN_DGTZ_IOLevel_TTL (1) or CAEN_DGTZ_IOLevel_NIM (0)
  ioLevel = (CAEN_DGTZ_IOLevel_t)config->v1740->ioLevel;
  CAEN_DGTZ_SetIOLevel(handle,ioLevel);
  CAENDecoder::checkError(retcod,"SetIOLevel",boardId);

  enableMask = 0xFF;
  CAEN_DGTZ_SetGroupEnableMask(handle,enableMask);
  CAENDecoder::checkError(retcod,"SetGroupEnableMask",boardId);

  CAEN_DGTZ_WriteRegister(handle,SAMPLE_REDUCTION_ADDR,
				   config->v1740->sampleReduction);
  CAENDecoder::checkError(retcod,"WriteSampleReduction",boardId);

  uint32_t readBack=0;
  CAEN_DGTZ_ReadRegister(handle,SAMPLE_REDUCTION_ADDR,&readBack);
  CAENDecoder::checkError(retcod,"WriteSampleReduction",boardId);
  if ( (int)readBack != config->v1740->sampleReduction )
  {
    time_t zeit;
    time(&zeit);
    std::cout << "*** " << std::ctime(&zeit) << " Error read " << readBack 
	      <<" from SampleReduction, expected " 
	      << config->v1740->sampleReduction << std::endl;
  }
}

void V1740Driver::configurePedestal()
{
  // BJR 10/29/14 comment out retcod as it is not used
  //CAEN_DGTZ_ErrorCode retcod;
  configureCommon();
  CAENDriver::configurePedestal();

  // Zero sums
  for ( int g=0; g<CAENDriver::MAX_GROUPS; g++)
  {
    gsum[g] = 0.0;
    gsumsq[g] = 0.0;
    nGSamples[g] = 0;
  }
}

void V1740Driver::readPedestals()
{
  std::ifstream pedFile;
  char line[256];

  bool found = false;

  pedFile.open("/daqdata/lariat_pedestal_group.dat");
  while ( !pedFile.eof() && !found)
  {
    pedFile >> line;
    if ( strcmp(line,"BoardId") == 0 )
    {
      pedFile >> line;
      int boardIdRead = atoi(line);
      if ( boardIdRead == boardIdRead )
      {
	pedFile >> line;
	pedFile >> line;
	found = true;
	for ( int g=0; g<CAENDriver::MAX_GROUPS; g++)
	{
	  int group;
	  double mean;
          pedFile >> line;
	  group = atoi(line);

	  pedFile >> line;
	  pedFile >> line;

	  mean = atof(line);
	  pedFile >> line;

	  offset[group] = (int)(mean * 16);
	}
      }
    }
  }
  pedFile.close();
}

void V1740Driver::configurePhysics()
{
  CAEN_DGTZ_ErrorCode retcod;

  configureCommon();
  CAENDriver::configurePhysics();

  if ( config->v1740->usePedestals )
  {
    readPedestals();
    for ( int group = 0; group<CAENDriver::MAX_GROUPS; group++)
    {
      retcod = CAEN_DGTZ_SetGroupDCOffset(handle,group,offset[group]);
      CAENDecoder::checkError(retcod,"SetGroupDCOffset",boardId);
    }
  }
}

void V1740Driver::readConfiguration()
{
  //CAEN_DGTZ_ErrorCode retcod;
  CAENDriver::readConfiguration();
  for (group=0; group<CAENDriver::MAX_GROUPS; group++)
  {
    CAEN_DGTZ_GetChannelGroupMask(handle,group,&channelMask[group]);
    CAEN_DGTZ_GetGroupTriggerThreshold(handle,group,&threshold[group]);
    CAEN_DGTZ_GetGroupDCOffset(handle,group,&offset[group]);
    CAEN_DGTZ_GetGroupSelfTrigger(handle,group,&groupSelfTriggerMode[group]);
  }
  CAEN_DGTZ_ReadRegister(handle,SAMPLE_REDUCTION_ADDR,
				  &sampleReduction);
  CAENDriver::readAcquisitionStatus();
}


void V1740Driver::print(std::ostream & os)
{
  CAENDriver::print(os);
  os << "  SampleReduction " << sampleReduction << std::endl;
  for ( group=0; group<CAENDriver::MAX_GROUPS; group++)
  {
    os << "\tGroup " << group << 
      "  Mask 0x" << std::hex << channelMask[group] << std::dec <<
      "  TrigThreshold " << threshold[group] << " DCOffset " << offset[group] <<
      " SelfTrig " << groupSelfTriggerMode[group] << " " <<  
      CAENDecoder::TriggerMode(groupSelfTriggerMode[group]) << std::endl;
  }
}

void V1740Driver::writePedestals(std::ostream *pedestalFile)
{
  double mean,rms; 
  char line[256];
  sprintf(line, "BoardNumber %d   BoardId %d  SerialNumber %d\n", 
	  boardChainNumber, 
	  config->v1740->boardId[boardChainNumber], boardInfo.SerialNumber);
  pedestalFile->write(line,std::strlen(line));
  for (int g=0; g<CAENDriver::MAX_GROUPS; g++)
  {
    if ( nSamples[g] > 0 )
    {
      mean = gsum[g] / (double)nGSamples[g];
      double n = (double)nGSamples[g];
      rms  = std::sqrt( gsumsq[g] / n - mean*mean );

      sprintf(line,"%3d %10d %8.4f %8.4f\n", g, nSamples[g], mean, rms);
      pedestalFile->write((char *)line,std::strlen(line));
    }
  }
}

