//
//  V1751Driver.cc  (W.Badgett)
//
//  Class to configure and readout one of a chain
//  of CAEN V1751 high speed digitizers for the Lariat PMTs
//

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <ptree.hpp>
#include "V1751Driver.h"
#include "CAENFragment.h"
#include "LariatConfiguration.h"
#include "LariatFragment.h"
#include "LariatReadout.h"
#include "CAENDecoder.h"

void V1751Driver::configureCommon()
{
  CAEN_DGTZ_ErrorCode retcod;
  uint32_t data, readBack;

  CAENDriver::configureCommon();

  // Turn off dual edge sampling
  retcod = CAEN_DGTZ_SetDESMode(handle,CAEN_DGTZ_DISABLE);
  CAENDecoder::checkError(retcod,"SetDESMode",boardId);

  if ( config->v1751->trigSource == TRIGGER_SOURCE_MAJORITY )
  {

    retcod = CAEN_DGTZ_ReadRegister(handle,TRIGGER_ADDR,&data);
    CAENDecoder::checkError(retcod,"ReadRegister TRIGGER_ADDR",boardId);

    // Set majority to 3
    data &= ~( 1 << 24 );
    data |=  ( 1 << 25 );
    data &= ~( 1 << 26 );

    // Set coincidence width to 8 ns
    data |=  ( 1 << 20 );
    data &= ~( 1 << 21 );
    data &= ~( 1 << 22 );
    data &= ~( 1 << 23 );

    retcod = CAEN_DGTZ_WriteRegister(handle,TRIGGER_ADDR,data);
    CAENDecoder::checkError(retcod,"WriteRegister TRIGGER_ADDR",boardId);

    retcod = CAEN_DGTZ_ReadRegister(handle,TRIGGER_ADDR,&readBack);
    CAENDecoder::checkError(retcod,"ReadRegister TRIGGER_ADDR",boardId);

    if ( data != readBack )
    {
      std::cout << CAENDecoder::zeit() << 
	" ***  Wrote TRIGGER_ADDR 0x%" << std::hex << data << " but read 0x" << readBack 
		<<std::endl;
    }
  }
  else if ( config->v1751->trigSource == TRIGGER_SOURCE_EXTERNAL )
  {
    retcod = CAEN_DGTZ_SetExtTriggerInputMode(handle, 
					      CAEN_DGTZ_TRGMODE_ACQ_ONLY); 
    CAENDecoder::checkError(retcod,"SetExtTriggerInputMode",boardId);
  }

  // Set 6th bit to 1 to allow negative thresholds

  retcod = CAEN_DGTZ_ReadRegister(handle,CHANNEL_CONFIG_ADDR,&data);
  CAENDecoder::checkError(retcod,"ReadRegister CHANNEL_CONFIG_ADDR",boardId);
  data |= (1<<6);

  retcod = CAEN_DGTZ_WriteRegister(handle,CHANNEL_CONFIG_ADDR,data);
  CAENDecoder::checkError(retcod,"WriteRegister CHANNEL_CONFIG_ADDR",boardId);

  readBack = 0;
  retcod = CAEN_DGTZ_ReadRegister(handle,CHANNEL_CONFIG_ADDR,&readBack);
  CAENDecoder::checkError(retcod,"ReadRegister CHANNEL_CONFIG_ADDR",boardId);
	     
  if ( data != readBack ) 
  {
    std::cout << CAENDecoder::zeit() << 
      " ***  Wrote CHANNEL_CONFIG_ADDR 0x" << std::hex << data << " but read 0x"
	      <<readBack << std::dec << std::endl;
  }

  // Single channel configurations

  
  int i=0;

  // BJR 10/29/14 comment out DESmultiplier here and on line 96 as it is never used
  //int DESmultiplier=1;
  if( config->v1751->desMode )
  { 
    i=1;
    //DESmultiplier=2;
  }
  
  //for now, I'm disregarding the DES mode and setting the mask assuming standard readout. A.S. 08/13/14
  int newMask=config->v1751->channelMask;
  
  std:: cout << "V1751 mask 0x" << std::hex << newMask << std::dec << std::endl;

  
  

 //  dcOffset[8]={zero_pos*0xFFFF+0.090*0xFFFF+0.1*fscal_corr,zero_pos*0xFFFF+0.080*0xFFFF+0.1*fscal_corr,zero_pos*0xFFFF+0.186*0xFFFF+0.9*fscal_corr,offs-corrections[3]*corr_fact,offs-corrections[4]*corr_fact,offs-corrections[5]*corr_fact,offs-corrections[6]*corr_fact,offs-corrections[7]*corr_fact};

  double zero_pos=0.4*0xFFFF;  // this is baseline at ~0 ADC; FS is 200mV instead of 1000mV and we want to be 512 ADC down from the middle. 
  double tick_size=0xFFFF/1024*0.2;  // again correcting for the smaller FS Size.
  
  
  double dcOffset[V1751Configuration::MAX_CHANNELS];
  //apply correction into proper ADC scale
  for(int ch=0;ch<V1751Configuration::MAX_CHANNELS;ch++)
   dcOffset[ch] =zero_pos+(double)(config->v1751->DCoffset[ch]*tick_size)+(double)(config->v1751->DCcorr[ch]*tick_size);
  
//cherenkov baseline val/
//int dcOffset[8]={zero_pos*0xFFFF+0.170*0xFFFF+0.1*fscal_corr,zero_pos*0xFFFF+0.080*0xFFFF+0.1*fscal_corr,zero_pos*0xFFFF+0.186*0xFFFF+0.9*fscal_corr,offs-corrections[3]*corr_fact,offs-corrections[4]*corr_fact,offs-corrections[5]*corr_fact,offs-corrections[6]*corr_fact,offs-corrections[7]*corr_fact};

  //signal should be at 0 volts, so trigger is at FSR[ADC] - dcOffset[ADC] - trig [ADC]
  //int value=1023-5*50-7;
 //  int value=daq_control.dc_offset-daq_control.trig_level;

  retcod = CAEN_DGTZ_SetChannelEnableMask(handle, newMask);
  for(i=0; i<V1751Configuration::MAX_CHANNELS; i++) 
  {
    if (newMask & (1<<i)) 
    {
      retcod = CAEN_DGTZ_SetChannelDCOffset(handle, i, dcOffset[i]);
      if( config->v1751->trigSource == TRIGGER_SOURCE_MAJORITY )
      {
        retcod = CAEN_DGTZ_SetChannelSelfTrigger(handle,CAEN_DGTZ_TRGMODE_ACQ_ONLY , (1<<i));
      }
      else if ( config->v1751->trigSource < TRIGGER_SOURCE_MAJORITY )
      {  
        //triggering on selected channel
        if( i == config->v1751->trigSource )
        {
  	 retcod = CAEN_DGTZ_SetChannelSelfTrigger(boardId,CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT , (1<<i));
        }
        retcod = CAEN_DGTZ_SetChannelTriggerThreshold(boardId, i, config->v1751->TrigThreshold[i]);
       // retcod = CAEN_DGTZ_SetTriggerPolarity(boardId, i, 1);
	std::cout << "Setting chan " << i << " threshold to: " << config->v1751->TrigThreshold[i]
		  << std::endl;	
      }
    }
  } // end channel for loop
}

void V1751Driver::configurePedestal()
{
  // BJR 10/29/15 comment out retcode as it is never used
  //CAEN_DGTZ_ErrorCode retcod;
  configureCommon();
  CAENDriver::configurePedestal();
}

void V1751Driver::readPedestals()
{
}

void V1751Driver::writePedestals(std::ostream *pedestalFile)
{
  double mean,rms; 
  char line[256];
  sprintf(line, "BoardNumber %d   BoardId %d  SerialNumber %d\n", 
	  boardChainNumber, 
	  config->v1751->boardId[boardChainNumber], boardInfo.SerialNumber);
  pedestalFile->write(line,std::strlen(line));
  for (int c=0; c<MAX_UINT16_CHANNEL_SIZE; c++)
  {
    if ( nSamples[c] > 0 )
    {
      mean = sum[c] / (double)nSamples[c];
      double n = (double)nSamples[c];
      rms  = std::sqrt( sumsq[c]/n - mean*mean );
      sprintf(line,"%3d %10d %8.4f %8.4f\n", c, nSamples[c], mean, rms);
      pedestalFile->write((char *)line,std::strlen(line));
    }
  }
}

void V1751Driver::configurePhysics()
{
  CAEN_DGTZ_ErrorCode retcod;

  configureCommon();
  CAENDriver::configurePhysics();

  if ( config->v1751->usePedestals )
  {
    //    CAENDriver::readPedestals();
    //    for ( int group = 0; group<CAENDriver::MAX_GROUPS; group++)
    // {
    //  retcod = CAEN_DGTZ_SetGroupDCOffset(handle,group,offset[group]);
    //  CAENDecoder::checkError(retcod,"SetGroupDCOffset",boardId);
    // }
  }

  // set the I/O lines to input
  // 0xBC = 1011 1100
  // See V1751 registers table 2.24 Front Panel I/O Control
  retcod = CAEN_DGTZ_WriteRegister(handle,FRONT_PANEL_ADDR,FRONT_PANEL_INPUT);
  CAENDecoder::checkError(retcod,"Set I/O lines",boardId);

}

void V1751Driver::readConfiguration()
{
  CAEN_DGTZ_ErrorCode retcod;
  CAENDriver::readConfiguration();

  retcod = CAEN_DGTZ_GetDESMode(handle,(CAEN_DGTZ_EnaDis_t *)&desMode);
  CAENDecoder::checkError(retcod,"GetDESMode",boardId);
}


void V1751Driver::print(std::ostream & os)
{
  CAENDriver::print(os);
  os << "  DESMode         " << desMode << " " << 
    CAENDecoder::EnaDisMode(desMode) <<  std::endl;
}

