#include "V2718Driver.h"


V2718Driver::V2718Driver(const LariatConfiguration* lariat, 
			 V2718Configuration* config, caen::vme_handle h_){
  h=h_;
  err=cvSuccess;
}


V2718Driver::~V2718Driver() 
{
  // This destructor assumes that the VME handle is passed
  // in from the CAENComm or CAENDigitizer libraries and that
  // they are doing the resource management.

}

void V2718Driver::check(CVErrorCodes error_code, const char* place){

  if( error_code != cvSuccess){
    std::string huh="V2718Driver : "; huh+=place; huh+=" : ";
    huh+=caen::decode_error(error_code);
    throw caen::caen_exception(huh,error_code);
  }

}

void V2718Driver::check(const char* place){
  check(err,place);
}


void V2718Driver::print(std::ostream& os){
  
  char swrel[20];
  err = CAENVME_SWRelease(swrel); check("swrel");
  char fwrel[20];
  err = CAENVME_BoardFWRelease(h,fwrel); check("fwrel");
  char drel[20];
  bzero(drel,20);
  err = CAENVME_DriverRelease(h,drel); check("drel");
  std::cout << " DriverRelease " << err << std::endl;
  using std::endl;
  os<<"========= V2718 ==========="<<endl;
  os<<"*Software Release : "<<swrel<<endl;
  os<<"*Firmware Release : "<<fwrel<<endl;
  os<<"*Driver Release   : "<<drel<<endl;

}



void V2718Driver::configureReadoutRegister(){

  // Clear Output 0  
  err= CAENVME_ClearOutputRegister(h,cvOut0Bit); 
  check("ClearOutputRegister");

  // configure input source 0
  // direct rather than inverted logic
  // LED is on during logical high
  err = CAENVME_SetInputConf(h,cvInput0,cvDirect,cvActiveHigh);
  check("SetInputConf");
  
  // set a v2718 pulser so that it starts on input 0
  // what it outputs is also configured 
  // CVTimeUnits timeunit=cvUnit25ns;
  CVTimeUnits timeunit=cvUnit104ms;
  // the combination of width=0 and pulseno=0 seems to produce an infinite pulse
  // couldn't manage to turn it off!
  unsigned char pulseno=1; // 
  unsigned char width=254; // time units
  unsigned char period=255; // time units
  CVIOSources start=cvInputSrc0;
  CVIOSources stop=cvManualSW;
  err=CAENVME_SetPulserConf(h,cvPulserA,period,width,timeunit,
			      pulseno,start,stop);
  check("SetPulserConf");

  // stop the pulser if it's on
  err=CAENVME_StopPulser(h,cvPulserA);
  check("StopPulser");

  // maybe we need to restart it??
  // does this act as an enable??
  err=CAENVME_StartPulser(h,cvPulserA);
  check("StartPulser");

  // now pipe the pulser output to one of the front panel outputs
  // the output will appear on LEMO 0
  // it is direct rather than inverted logic 
  // the LED should be active high
  // cvMiscSignals = Take the signal from the list of misc signals
  // for output 0, it's "misc signal" is the output of pulser 0
  err = CAENVME_SetOutputConf(h,cvOutput0,cvDirect,
			      cvActiveHigh,cvMiscSignals);
  check("SetOutputConf");

}

bool V2718Driver::spill()
{
  bool spill = false;
  uint32_t data = 0;

  err = CAENVME_ReadRegister(h,cvInputReg,&data);
  check("ReadRegister in spill()");
  spill = data & cvIn0Bit;

  return(spill);
}


bool V2718Driver::pollReadoutRegister(int ntimes, useconds_t time_sep){
  bool pulser_active=false;
  uint32_t data=0xFFFF;

  for(int i=0; i<ntimes; i++){
    err=CAENVME_ReadRegister(h,cvInputReg,&data);      
    check("ReadRegister in pollReadoutRegister");
    pulser_active=data&cvPulsAOutBit;
    if(pulser_active) break;
    else {
      usleep(time_sep);
    }
  }
  return pulser_active;
  
}

void V2718Driver::resetReadoutRegister(){
  // acts as a disable
  err=CAENVME_StopPulser(h,cvPulserA); 
  check("StopPulser");
  // acts as an enable since the pulser
  // is configured to start when a signal is seen on input 0
  err=CAENVME_StartPulser(h,cvPulserA);
  check("StartPulser");
}

