//
//  WUTDriver.cc  (W.Badgett)
//
//  Base class to configure and readout one 
//   WUT digitizer (Wave Union TDC from Jinyuan)
//

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include <ptree.hpp>
#include <time.h>
#include <stdexcept>
#include "WUTDriver.h"
#include "WUTFragment.h"
#include "LariatConfiguration.h"
#include "LariatFragment.h"
#include "LariatReadout.h"

WUTDriver::~WUTDriver()
{
  for ( uint i=0; i<N_BUFFERS; i++)
  { if ( buffer[i] != NULL ) { free(buffer[i]);}}

}

WUTDriver::WUTDriver(LariatConfiguration * config_,
		     WUTConfiguration    * wut_,
		     int                   handle_)
{
  CVErrorCodes retcod;

  config            = config_;
  wut               = wut_;
  handle            = handle_;
  base              = wut->baseAddress;
  firmwareVersion   = 0;
  thresholdRegister = 0;
  patternWord       = 0;
  csr               = 0;

  for ( uint i=0; i<N_BUFFERS; i++)
  {
    buffer[i] = (uint32_t *)malloc(MAX_WORDS*sizeof(uint32_t));
    memset(buffer[i],0xAA,MAX_WORDS*sizeof(uint32_t));
  }

  for ( uint i=0; i<MAX_WORDS; i++)
  {
    page[0][i] = DATA0_ADDR | (i*4);
    page[1][i] = DATA1_ADDR | (i*4);
    am[i]      = cvA24_U_DATA;
    dw[i]      = cvD32;
  }



  thresholdMask = ( ( wut->threshold[3] << THRESHOLD3_S ) | 
		    ( wut->threshold[1] << THRESHOLD1_S ) );

  retcod = CAENVME_WriteCycle(handle,base+THRESHOLD_ADDR,(void *)&thresholdMask,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT Write32 Threshold");

  uint32_t thresholdRead=0xFFFFFFFF;
  retcod = CAENVME_ReadCycle(handle,base+THRESHOLD_ADDR,(void *)&thresholdRead,
			     cvA24_U_DATA,cvD32);
  if ( thresholdRead != thresholdMask )
  {
    std::cout << CAENDecoder::zeit() << 
      " WUT threshold write 0x" << std::hex << thresholdMask << " read 0x" <<
      thresholdRead << std::dec << std::endl;
  }

  csrMask = 0;
  if ( wut->externalTrigger ) 
  { csrMask |= (TRIGGER_ENABLE_M | TRIGGER_NIM0_M);}

  csrMask |= ( wut->calibrationDelay << CALIB_DELAY_S);
  if ( wut->useNim1 ) 
  {
    csrMask |= USE_NIM1_M ;
  }
  if ( wut->disableSeparator )
  {
    csrMask |= DISABLE_SEPARATOR_M;
  }
  retcod = CAENVME_WriteCycle(handle,base+CSR_ADDR,(void *)&csrMask,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT Write32 CSR");

  uint32_t csrRead=0xFFFFFFFF;
  retcod = CAENVME_ReadCycle(handle,base+CSR_ADDR,(void *)&csrRead,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT Read32 CSR");
  csrRead &= CONTROL_M;
  if ( csrRead != csrMask )
  {
    std::cout << CAENDecoder::zeit() << " WUT CSR Wrote " << 
      csrMask << " Read " << csrRead << std::endl;
  }

  retcod = CAENVME_WriteCycle(handle,base+DISABLE_ADDR,(void *)&wut->disableMask,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT Write32 Disable");

  uint32_t readBack = 0xFFFFFFFF;
  retcod = CAENVME_ReadCycle(handle,base+DISABLE_ADDR,(void *)&readBack,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT Read32 Disable");
  if ( readBack != wut->disableMask )
  {
    std::cout << CAENDecoder::zeit() << " WUT DisableMask Wrote " << 
      wut->disableMask << " Read " << readBack << std::endl;
  }

  readConfiguration();
}

void WUTDriver::decodeCSR()
{
  tdcChannel    = (( csr & TDC_CHANNEL_M ) >> TDC_CHANNEL_S );
  changePage    = (( csr & CHANGE_PAGE_M )    != 0);
  freezeStatus  = (( csr & FREEZE_STATUS_M )  != 0);
  triggerEnable = (( csr & TRIGGER_ENABLE_M ) != 0);
  triggerNim0   = (( csr & TRIGGER_NIM0_M )   != 0);
  triggerVme    = (( csr & TRIGGER_VME_M )    != 0);
  useNim1       = (( csr & USE_NIM1_M )       != 0 );
  calibVme      = (( csr & CALIB_VME_M )      != 0 );
  calibDelay    = (( csr & CALIB_DELAY_M ) >> CALIB_DELAY_S);
  bufferActive  = (( csr & BUFFER_PAGE_M)>>BUFFER_PAGE_S);
  bufferStoring = (( csr & BUFFER_STORING_M)!=0);
  wordCount     = (( csr & WORD_COUNT_M)>>WORD_COUNT_S);
}

uint32_t WUTDriver::readAndFlip()
{
  CVErrorCodes retcod;
  uint32_t localCsr;

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&localCsr,
			     cvA24_U_DATA,cvD32);
  localCsr &= 0xFFFF;
  CAENDecoder::vmeError(retcod,"WUT readAndFlip CSR Read");

  // Toggle the page flipper
  uint32_t flippy = localCsr | CHANGE_PAGE_M;

  retcod = CAENVME_WriteCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&flippy,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readAndFlip CSR Write");

  retcod = CAENVME_WriteCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&flippy,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readAndFlip CSR Write");

  retcod = CAENVME_WriteCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&localCsr,
			      cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readAndFlip CSR Write");

  // Sleep, my pretty
  struct timespec sleepyTime;
  sleepyTime.tv_sec  = 0;
  sleepyTime.tv_nsec = config->wut->hitSleep;
  nanosleep(&sleepyTime,NULL);

  // Read word counts
  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&localCsr,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readAndFlip CSR Read");
  csr = localCsr;
  wordCount = ((localCsr & WORD_COUNT_M ) >> WORD_COUNT_S);
  pageRead  = ( ( localCsr & BUFFER_PAGE_M ) >> BUFFER_PAGE_S);

  if (pageRead == 0 )
  {
    return((uint32_t)DATA0_ADDR);
  }
  else
  {
    return((uint32_t)DATA1_ADDR);
  }
}

void WUTDriver::readConfiguration()
{
  CVErrorCodes retcod;
  //uint32_t data;

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)FIRMWARE_ADDR,(void *)&firmwareVersion,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readConfiguratino Firmware");

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)THRESHOLD_ADDR,(void *)&thresholdRegister,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readConfiguratino Threshold");
  threshold1 = (thresholdRegister >> THRESHOLD1_S ) & THRESHOLD_M;
  threshold3 = (thresholdRegister >> THRESHOLD3_S ) & THRESHOLD_M;

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)PATTERN_ADDR,(void *)&patternWord,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readConfiguratino Patter");
  if ( patternWord != (uint32_t)WUTConfiguration::EXPECTED_PATTERN_WORD )
  {
    std::cout << CAENDecoder::zeit() << 
      " WUT patterWord expected 0x" << std::hex << 
      WUTConfiguration::EXPECTED_PATTERN_WORD << " found 0x" <<
      patternWord << std::dec << std::endl;
  }

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)CSR_ADDR,(void *)&csr,cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readConfiguratino CSR");
  decodeCSR();

  retcod = CAENVME_ReadCycle(handle,base+(uint32_t)DISABLE_ADDR,(void *)&disableMask,
			     cvA24_U_DATA,cvD32);
  CAENDecoder::vmeError(retcod,"WUT readConfiguratino Disable");
}


void WUTDriver::print(std::ostream & os)
{
  os << "WUT Device at link " << wut->link << 
    " board " << wut->boardNumber << 
    " VME base 0x" << std::hex << wut->baseAddress << std::endl; 
  os << "  FirmwareVersion     0x%" << firmwareVersion << std::endl;
  os << "  ThresholdRegister   0x" << thresholdRegister << "   Vth1 0x" 
     << threshold1 << " Vth3 0x" << threshold3 << std::endl;
  os << "  PatternWord         0x" << patternWord << std::endl;
  os << "  DisableMask         0x" << disableMask << std::endl;
  os << "  CSR                 0x" << csr << std::dec << std::endl;
  os << "    ChangePage " << changePage <<
    "  FreezeStatus " << freezeStatus << 
    "  TdcChannel " << tdcChannel << std::endl; 
  os << "    Trigger: Enable " << triggerEnable << 
    "  Nim0 " << triggerNim0 << 
    "  Vme " << triggerVme << std::endl;
  os << "    Calib:   Vme " << calibVme << 
    "  Delay " << calibDelay << "    UseNim1: " << useNim1 << std::endl;
  os << "    Buffer:  Active " << bufferActive << 
    "  Storing " << bufferStoring <<
    "  WordCount: " <<  wordCount << std::endl;
  os << "  PageRead            0x" << std::hex << pageRead << std::endl;
  os << "  WordCount           0x" << wordCount << std::dec << std::endl;
}


void WUTDriver::readEvent(std::ostream *dataFile, int pollCount)
{
  CVErrorCodes retcod;
  struct timespec before,after;

  clock_gettime(CLOCK_REALTIME, &before);

  uint32_t address = readAndFlip();

  int bufferNumber = 0;
  if ( address == DATA1_ADDR ) { bufferNumber = 1;}

  int nBytes=0; 

  if ( config->wut->enableBLT )
  {
    if ( wordCount > 0 )
    {
      retcod = CAENVME_BLTReadCycle(handle,
				    base+address,
				    buffer[bufferNumber],
				    wordCount*sizeof(wordCount),
				    cvA24_U_DATA,cvD32,
				    &nBytes);
      CAENDecoder::vmeError(retcod,"WUT readEvent BLT data");
      if ( wordCount*sizeof(wordCount) != (uint32_t)nBytes )
      {
	std::cout << CAENDecoder::zeit() << " WUT BLT expected  " << 
	  wordCount*sizeof(uint32_t) << " bytes, returned  " <<
	  nBytes << " bytes" << std::endl;
      }
    }
  }
  else
  {
    retcod = CAENVME_MultiRead(handle,
			       base+page[bufferNumber],
			       buffer[bufferNumber],
			       wordCount,
			       am,dw,
			       retcods);
    CAENDecoder::vmeError(retcod,"WUT readEvent data");

    // Each transaction could potentially have its own error
    for ( uint h=0; h<wordCount; h++)
    {
      char blah [8];
      sprintf(blah,"%d ", h);
      CAENDecoder::vmeError(retcods[h],blah);
    }
  }


  clock_gettime(CLOCK_REALTIME, &after);
  if ( wut->debugLevel >= 6 )
  {
    unsigned long long w = before.tv_sec * 1000000000 + before.tv_nsec;
    unsigned long long d =  after.tv_sec * 1000000000 + after.tv_nsec;
    unsigned long long delta = d - w;
    
    std::cout << "Got WUT event of " << wordCount << " in " << 
      ( delta / 1000000000 ) << " , " << ( delta % 1000000000 )
	      << std::endl;
  }	

  uint32_t w=0;
  while ( w < wordCount )
  {
    if ( buffer[bufferNumber][w] != 0 )
    {
      if ( ( buffer[bufferNumber][w] & MARKER_M ) == TIME_FRAME_MARKER ) // new timeframe sighted
      {
	WUTFragment frag;
	frag.header.fragmentSize = sizeof(frag.header);
	frag.header.fragmentType = LariatFragment::FRAGMENT_TYPE_WUT_HIT_HEADER;
	frag.header.pollCount    = pollCount;
	frag.header.spillCounter = 1;

	// Header time in units of 16 us, except for four lower bits contained in hit words
	frag.header.timeHeader = ((buffer[bufferNumber][w]) & TCY_HEADER_M)<< TCY_HEADER_S;

	w++;
	int iHit=0;
	uint32_t tcyHit = 0;
	if ( w < wordCount ) // now read hits
	{
	  while ( ( ( buffer[bufferNumber][w] & MARKER_M ) == HIT_MARKER ) && ( w < wordCount )) 
	  {
	    // Found data from a hit
	    frag.header.fragmentSize += sizeof(struct WUTFragment::WutHit);
	    WUTFragment::WutHit_t hit;

	    hit.channel = (buffer[bufferNumber][w] >> HIT_CHANNEL_S) & HIT_CHANNEL_M;

	    if(iHit==0)
	    { 
	      tcyHit = ( buffer[bufferNumber][w] & TCY_HIT_M  ) >> TCY_HIT_S;
	      frag.header.timeHeader = frag.header.timeHeader | tcyHit;
	    }
	    else
	    {
	      // The TCY_HIT value should be identical for all hits -- check it
	      uint32_t thisTcyHit = ( buffer[bufferNumber][w] & TCY_HIT_M  ) >> TCY_HIT_S;
	      if ( tcyHit != thisTcyHit )
	      {
		std::cout << 
		  "TCY_HIT mismatch hit 0 says 0x" << std::hex << tcyHit << 
		  ", hit " << std::dec << iHit << " says 0x%" << 
		  std::hex << thisTcyHit << std::dec << std::endl;
	      }
	    }

	    hit.timeBin = buffer[bufferNumber][w] & HIT_TIME_M;
	    frag.hits.push_back(hit);
	    iHit++;
	    w++;
	  }
	}

	frag.header.nHits = iHit;
	if ( config->wut->debugLevel >= 7 ) { frag.print(std::cout); std::cout.flush();}

	// Fragment is formed, now write it out
	frag.streamToBuffer(*dataFile);
      }
      else
      {
	// Quietly ignore?
	std::cout << 
	  "*** WUTDriver::readEvent Found unexpected data 0x" << std::hex << 
	  buffer[bufferNumber][w] << " at 0x" << 
	  w << " buffer " << std::dec << bufferNumber << std::endl;
	w++;
      }
    }
    else { w++;} // Skip optional zeroes
  }
}
