#include <iostream>
#include <fstream>
#include <cassert>
#include <cstdio>
#include <vector>
#include <string>

#include "LariatFragment.h"
//#include "TDCFragments.h"

// ROOT includes
#include <TROOT.h>
#include <TFile.h>
#include <TTree.h>
#include <TH1D.h>

std::string remove_ext(std::string full_name);

void parse(std::string file_path, std::string delimiter,
           std::vector<std::string> & token_vector);

int main(int argc, char * argv[]) 
{

  std::string file_path = argv[1];

  std::vector<std::string> run_token_vector;
  parse(file_path, "_", run_token_vector);

  std::cout << "Run:    " << run_token_vector.end()[-2] << std::endl;

  gROOT->ProcessLine("#include <vector>");

  std::string output_file_path = "exp_run_" +
      run_token_vector.end()[-2] + ".root";

  TFile * file = new TFile(output_file_path.c_str(), "RECREATE");
  TTree * tree = new TTree("SpillTree", "");

  int SpillID;
  int EventID;
  long EventTime;
  int HitID;
  int HitTDCID;
  int HitWireID;
  int HitTime;

  // http://root.cern.ch/root/html/TBranch.html#TBranch:TBranch@2
  tree->Branch("SpillID", &SpillID, "SpillID/I");
  tree->Branch("EventID", &EventID, "EventID/I");
  tree->Branch("EventTime", &EventTime, "EventTime/L");
  tree->Branch("HitID", &HitID, "HitID/I");
  tree->Branch("HitTDCID", &HitTDCID, "HitTDCID/I");
  tree->Branch("HitWireID", &HitWireID, "HitWireID/I");
  tree->Branch("HitTime", &HitTime, "HitTime/I");

  assert(argc >= 1);
  //std::ifstream in(argv[1]);

  std::cout << "TDCFragment::MAX_CHAMBERS: " << TDCFragment::MAX_CHAMBERS
            << std::endl;
  std::cout << "TDCFragment::MAX_TDCS: " << TDCFragment::MAX_TDCS
            << std::endl;
  std::cout << "TDCFragment::TDCS_PER_CHAMBER: "
            << TDCFragment::TDCS_PER_CHAMBER << std::endl;
  std::cout << "TDCFragment::N_CHANNELS: " << TDCFragment::N_CHANNELS
            << std::endl;

  for (int file_index = 1; file_index < argc; ++file_index) {

    std::ifstream in(argv[file_index]);
    std::vector<std::string> token_vector;
    parse(argv[file_index], "_", token_vector);

    //spill = (uint16_t) atoi(token_vector.end()[-1].c_str());
    //std::cout << "Spill:  " << spill << std::endl;

    SpillID = (int) atoi(token_vector.end()[-1].c_str());
    std::cout << "Spill:  " << SpillID << std::endl;

    LariatFragment spill_data;

    uint32_t readBytes = spill_data.fillFromBuffer(in);
    std::cout << "Read " << readBytes << " bytes from " << argv[file_index]
              << std::endl;

    const int numberTdcFrags = spill_data.tdcFrags.size();
    std::cout << "Found " << numberTdcFrags << " TDC fragments" << std::endl;

    std::cout << "Looking at TDC fragments..." << std::endl;

    for (int i = 0; i < numberTdcFrags; ++i) {

      TDCFragment & tdcFrag = spill_data.tdcFrags[i];

      std::vector< std::vector<TDCFragment::TdcEventData> > &
          tdcEvents = tdcFrag.tdcEvents;

      std::cout << "tdcEvents.size(): " << tdcEvents.size() << std::endl;
      tdcFrag.print();

      for (uint j = 0; j < tdcEvents.size(); ++j) {

        EventID = j;

        if (tdcFrag.controllerHeader.nTDCs != tdcEvents[j].size()) {
          std::cout << "*** Fatal nTDCs mismatch: " << tdcEvents[j].size()
                    << " != " << tdcFrag.controllerHeader.nTDCs << std::endl;
        }

        //uint16_t total_hits = 0;

        for (int tdc_index = 0; tdc_index < TDCFragment::MAX_TDCS;
             ++tdc_index) {
          TDCFragment::TdcEventData data = tdcEvents[j].at(tdc_index);

          //controller_time_stamp = data.tdcEventHeader.controllerTimeStamp;
          //trigger_time_stamp = data.tdcEventHeader.tdcTimeStamp;

          uint64_t trigger =
              ((uint64_t) data.tdcEventHeader.tdcTimeStamp << 3) |
              ((uint64_t) data.tdcEventHeader.controllerTimeStamp &0x7);
          //double trigger_time = ((double) trigger) / 1.177E09;
          EventTime = (long) trigger;


          //if (data.tdcEventHeader.nHits == 0) continue;

          //tdc.push_back(tdc_index + 1);
          //number_hits.push_back(data.tdcEventHeader.nHits);
          //total_hits += data.tdcEventHeader.nHits;

          for (uint hit_index = 0; hit_index < data.tdcHits.size();
               ++hit_index) {
            TDCFragment::TdcHit & hit = data.tdcHits[hit_index];
            //hit_channel.push_back((uint16_t) hit.channel);
            //hit_time_bin.push_back(hit.timeBin);

            HitID = hit_index + 1;
            HitTDCID = tdc_index + 1;
            HitWireID = (int) hit.channel;
            HitTime = (int) hit.timeBin;

            tree->Fill();
          }
        }
      }
    }
  }

  file->Write();
  delete file;

  return 0;

}

std::string remove_ext(std::string full_name) {
  int last_index = full_name.find_last_of(".");
  std::string raw_name = full_name.substr(0, last_index);
  return raw_name;
}

void parse(std::string file_path, std::string delimiter,
           std::vector<std::string> & token_vector) {
  size_t pos = 0;
  std::string token;
  while ((pos = file_path.find(delimiter)) != std::string::npos) {
    token = file_path.substr(0, pos);
    token_vector.push_back(token);
    file_path.erase(0, pos + delimiter.length());
  }
  token_vector.push_back(remove_ext(file_path));
}

